"""Test the module `aws_orga_deployer.execution`."""

import shutil
import tempfile
import unittest
from unittest.mock import patch

import boto3
from moto import mock_cloudformation, mock_organizations, mock_s3, mock_ssm, mock_sts
from moto.server import ThreadedMotoServer

from aws_orga_deployer import modules
from aws_orga_deployer.execution import Executor
from aws_orga_deployer.orga import OrgaParser
from aws_orga_deployer.package import Package
from tests import mock
from tests.mock import mock_others
from tests.utils import update_cli_filters


class TestExecutor(unittest.TestCase):
    """Test the class Executor."""

    @mock_sts
    @mock_others
    def setUp(self):
        """Create resources that must exist for all tests."""
        mock.mock_cli_arguments(package_filename="package2.yaml")
        # Create a temporary folder for the test
        self.temp_dir = tempfile.mkdtemp()
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"temp_dir": self.temp_dir}),
        ):
            # We enable moto threaded server for CloudFormation and Terraform
            # testing as they are executed in separate subprocesses
            self.server = ThreadedMotoServer(verbose=False)
            self.server.start()
            # Enable mocking for S3 and Organizations in all tests without
            # having to add declarators to each test function
            self.mock_s3 = mock_s3()
            self.mock_s3.start()
            mock.create_fake_bucket()
            # Load modules
            modules.load_modules()
            # Enable persistent mocking for Organizations
            self.mock_organizations = mock_organizations()
            self.mock_organizations.start()
            # Create a fake AWS organization
            self.fake_orga = mock.FakeOrgaParser()
            # Create an OrgaParser object that queries the fake organization.
            # A Package object must be created first to load package configuration
            # settings, but it is unused after
            Package()
            self.orga = OrgaParser()

    def tearDown(self):
        """Disable mocking and delete common resources."""
        # Disable mocking
        self.mock_s3.stop()
        self.mock_organizations.stop()
        self.server.stop()
        # Delete the temporary folder
        shutil.rmtree(self.temp_dir)

    def test_python(self):
        """Test deployments with Python engine."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters(
                {"temp_dir": self.temp_dir, "include_modules": ["python1"]}
            ),
        ):
            package = Package()
            package.full_init(self.orga)
            executor = Executor(package)
            executor.run()
            export = package.export_results()
            self.assertEqual(len(export["Completed"]["Create"]), 1)
            # The detailed results are expected to contain the default AWS
            # account ID of moto (123456789012)
            self.assertEqual(
                export["Completed"]["Create"][0]["DetailedResults"]["AccountId"],
                "123456789012",
            )

    @mock_cloudformation
    @mock_ssm
    def test_cloudformation_preview(self):
        """Test to preview changes with the CloudFormation module."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({
                "temp_dir": self.temp_dir,
                "command": "preview",
                "include_modules": ["cloudformation1"],
            }),
        ):
            package = Package()
            package.full_init(self.orga)
            executor = Executor(package)
            executor.run()
            export = package.export_results()
            self.assertEqual(len(export["Completed"]["Create"]), 1)

    @mock_ssm
    @mock_cloudformation
    def test_cloudformation_create_destroy(self):
        """Test to create and destroy deployments with the CloudFormation
        module. Update cannot be tested because of a bug in moto:
        `create_change_set` fails with action is `UPDATE`. Also, moto does not
        return the stack outputs when a stack is created from a change set, and
        therefore cannot be tested.
        """
        ssm_client = boto3.client("ssm", region_name="eu-west-1")
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters(
                {"temp_dir": self.temp_dir, "include_modules": ["cloudformation1"]}
            ),
        ):
            # Create the deployment and check that one deployment was created
            # and the SSM parameter does exist
            package1 = Package()
            package1.full_init(self.orga)
            executor1 = Executor(package1)
            executor1.run()
            export1 = package1.export_results()
            self.assertEqual(len(export1["Completed"]["Create"]), 1)
            package1.save()
            parameter = ssm_client.get_parameter(Name="test")["Parameter"]
            self.assertEqual(parameter["Value"], "valueCloudFormation1")
            # Remove the deployments from the package, which requires to
            # destroy the existing deployment. Then, check that the SSM
            # parameter doesn't exist anymore
            package2 = Package()
            package2.package["Modules"]["cloudformation1"]["Deployments"] = []
            package2.full_init(self.orga)
            executor2 = Executor(package2)
            executor2.run()
            export2 = package2.export_results()
            self.assertEqual(len(export2["Completed"]["Destroy"]), 1)
            with self.assertRaises(ssm_client.exceptions.ParameterNotFound):
                parameter = ssm_client.get_parameter(Name="test")["Parameter"]

    @mock_ssm
    def test_terraform_preview(self):
        """Test to preview changes with the Terraform module."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({
                "temp_dir": self.temp_dir,
                "command": "preview",
                "include_modules": ["terraform1"],
            }),
        ):
            package = Package()
            package.full_init(self.orga)
            executor = Executor(package)
            executor.run()
            export = package.export_results()
            self.assertEqual(len(export["Completed"]["Create"]), 1)

    @mock_ssm
    def test_terraform_create_update_destroy(self):
        """Test to create, update and destroy deployments with the Terraform
        module.
        """
        ssm_client = boto3.client("ssm", region_name="eu-west-1")
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters(
                {"temp_dir": self.temp_dir, "include_modules": ["terraform1"]}
            ),
        ):
            # Create the deployment and check that one deployment was created
            # and the SSM parameter does exist
            package1 = Package()
            package1.full_init(self.orga)
            executor1 = Executor(package1)
            executor1.run()
            export1 = package1.export_results()
            package1.save()
            self.assertEqual(len(export1["Completed"]["Create"]), 1)
            self.assertIn(
                "SSMParameterARN", export1["Completed"]["Create"][0]["Outputs"]
            )
            parameter = ssm_client.get_parameter(Name="test")["Parameter"]
            self.assertEqual(parameter["Value"], "valueTerraform1")
            # Update the deployment by changing the value of the variable
            package2 = Package()
            vars_module = package2.package["Modules"]["terraform1"]["Variables"]
            vars_module["varTerraform1"] = "newValue"
            package2.full_init(self.orga)
            executor2 = Executor(package2)
            executor2.run()
            export2 = package2.export_results()
            package2.save()
            self.assertEqual(len(export2["Completed"]["Update"]), 1)
            parameter = ssm_client.get_parameter(Name="test")["Parameter"]
            self.assertEqual(parameter["Value"], "newValue")
            # Remove the deployments from the package, which requires to
            # destroy the existing deployment. Then, check that the SSM
            # parameter doesn't exist anymore
            package3 = Package()
            package3.package["Modules"]["terraform1"]["Deployments"] = []
            package3.full_init(self.orga)
            executor3 = Executor(package3)
            executor3.run()
            export3 = package3.export_results()
            self.assertEqual(len(export3["Completed"]["Destroy"]), 1)
            with self.assertRaises(ssm_client.exceptions.ParameterNotFound):
                parameter = ssm_client.get_parameter(Name="test")["Parameter"]



"""Wrapper of Python modules that is executed in a subprocess."""

import importlib.util
import sys

from aws_orga_deployer.engines.wrappers import base


def main():
    """Main function."""
    # Read the inputs from the JSON file `input.json`
    inputs = base.read_wrapper_inputs()
    # Import the module Python script
    module_file = sys.argv[1]
    spec = importlib.util.spec_from_file_location("main", module_file)
    main_module = importlib.util.module_from_spec(spec)
    sys.modules["main"] = main_module
    spec.loader.exec_module(main_module)
    # Execute the main function of the module
    made_changes, result, detailed_results, outputs = main_module.main(
        inputs.module,
        inputs.account_id,
        inputs.region,
        inputs.command,
        inputs.action,
        inputs.variables,
        inputs.module_config,
        inputs.module_dir,
        inputs.deployment_cache_dir,
        inputs.engine_cache_dir,
    )
    # Write the outputs to a file as a JSON file
    base.write_wrapper_outputs(made_changes, result, detailed_results, outputs)


if __name__ == "__main__":
    main()



"""Execute the steps of a deployment package."""

import logging
import os
import shutil
import signal
import subprocess
import time
from datetime import datetime
from threading import Event, Lock, Thread
from typing import Any, Dict

from aws_orga_deployer import config, utils
from aws_orga_deployer.engines import ENGINES
from aws_orga_deployer.package import Package, graph
from aws_orga_deployer.package.store import ModuleAccountRegionKey

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class SubprocessError(Exception):
    """Exception to raise if an error occurs when running subprocesses."""


class Executor:
    """Class that is used to execute the steps of a deployment package using
    the modules.

    Attributes:
        package: Current package.
        nb_times_sigint: Number of times CTRL+C was pressed.
        stop_workers: Event object used to inform worker threads that they must
            stop after completing the current step.
        send_sigint: Event object used to inform worker threads that they must
            send a SIGINT signal to their subprocesses to quit gracefully.
        send_sigint: Event object used to inform worker threads that they must
            send a SIGTERM signal to their subprocesses to force quit.
        aws_temp_credentials: Store AWS temporary credentials retrieved for each
            IAM role and when they were obtained. Avoid to do an AssumeRole for
            every step.
        lock_aws_temp_credentials: Lock object to avoid multiple threads to
            retrieve AWS temporary credentials at the same time.
        lock_next: Lock object to avoid multiple threads to retrieve the next
            step at the same time.
        engines_cache_dir: Path to the folder that contains engines cache
        deployments_cache_dir: Path to the folder that contains deployments cache
        root_logs_dir: Root path to the logs of this run
    """

    package: Package
    nb_times_sigint: int
    stop_workers: Event
    send_sigint: Event
    send_sigterm: Event
    aws_temp_credentials: Dict[str, Any]
    lock_aws_temp_credentials: Lock
    lock_next: Lock
    engines_cache_dir: Dict[str, str]
    deployments_cache_dir: str
    root_logs_dir: str

    def __init__(self, package: Package) -> None:
        self.package = package
        self.nb_times_sigint = 0
        self.stop_workers = Event()
        self.send_sigint = Event()
        self.send_sigterm = Event()
        self.aws_temp_credentials = {}
        self.lock_aws_temp_credentials = Lock()
        self.lock_next = Lock()
        self._create_temp_folders()

    def _create_temp_folders(self) -> None:
        """Create the temporary folders in the local disk to store cache files
        and logs.
        """
        cache_dir = os.path.join(config.CLI["temp_dir"], "cache")
        # Create one cache folder for each engine (cache/engines/engine)
        self.engines_cache_dir = {}
        for engine in ENGINES:
            engine_dir = os.path.join(cache_dir, "engines", engine)
            self.engines_cache_dir[engine] = engine_dir
            if not os.path.isdir(engine_dir):
                os.makedirs(engine_dir)
        # Delete deployment cache folders if exist (cache/deployments)
        self.deployments_cache_dir = os.path.join(cache_dir, "deployments")
        if os.path.exists(self.deployments_cache_dir):
            shutil.rmtree(self.deployments_cache_dir)
        # Evaluate if deployment cache must persist after this run
        self.delete_deployment_cache = (
            config.CLI.get("keep_deployment_cache", False) is False
        )
        # Create the logs folder (outputs/YYYYMMDD-HHMMSS)
        current_date = datetime.utcnow().strftime("%Y%m%d-%H%M%S")
        self.root_logs_dir = os.path.join(config.CLI["temp_dir"], "logs", current_date)
        if not os.path.isdir(self.root_logs_dir):
            os.makedirs(self.root_logs_dir)

    def run(self) -> None:
        """Function to call to launch the execution of steps."""
        # Catch SIGINT signals to be able to stop gracefully the processing of
        # steps
        signal.signal(signal.SIGINT, self._catch_sigint)
        # Launch concurrent threads to execute the steps
        nb_threads = config.PACKAGE.get(
            "ConcurrentWorkers", config.DEFAULT_CONCURRENT_WORKERS
        )
        threads = []
        for worker_id in range(nb_threads):
            thread = Thread(
                target=self.worker,
                daemon=True,
                name=f"Worker{worker_id}",
            )
            thread.start()
            threads.append(thread)
        # Wait until the threads complete
        for thread in threads:
            thread.join()
        # Save the current state store before exiting
        self.package.save()
        # Delete the deployments cache root folder if needed
        if self.delete_deployment_cache:
            shutil.rmtree(self.deployments_cache_dir, ignore_errors=True)
        # Stop catching SIGINT signals
        signal.signal(signal.SIGINT, Executor._stop_catching_sigint)

    def worker(self) -> None:
        """Main function executed by concurrent workers."""

        LOGGER.debug("Starting worker")
        while True:
            # Stop the worker is `stop_workers` is set
            if self.stop_workers.is_set():
                break
            # Get the next deployment to process. Use a lock to avoid
            # multiple workers to update the package graph at the same time
            # pylint: disable=bare-except
            # pylint: disable=broad-exception-caught
            try:
                with self.lock_next:
                    key, action, nb_attempts, max_attempts = self.package.next()
                    LOGGER.info(
                        "%s Starting to %s (Attempt %s/%s)",
                        key,
                        action,
                        nb_attempts,
                        max_attempts,
                    )
            # Wait one second if there are still pending deployments but they
            # are all waiting for dependencies to complete
            except graph.NoProcessableStep:
                time.sleep(1)
                continue
            # Stop the worker if there are no more pending deployments
            except graph.NoMorePendingStep:
                break
            # Stop the worker if `next` failed
            except:
                LOGGER.error(
                    "Worker failed to get the next deployment to process",
                    exc_info=config.CLI["debug"],
                )
                break
            # If the requested command is "update-hash", attempt to update the
            # module hash and move to the next deployment to process
            if config.CLI["command"] == "update-hash":
                hash_changed = self.package.update_hash(key)
                if hash_changed:
                    LOGGER.info("%s Updated the value of the module hash", key)
                else:
                    LOGGER.info("%s No action needed", key)
                continue
            # Use this variable to indicate where the step fails in the
            # detailed error logs
            section_that_failed = "other"
            # Determine the name of the deployment cache
            deployment_cache_dir = os.path.join(
                self.deployments_cache_dir, key.module, key.account_id, key.region
            )
            # Catch any exceptions to prevent the worker thread from being
            # interrupted if the module fails
            try:
                # Measure the step execution time
                start_time = datetime.now()
                # Create the deployment cache dir
                os.makedirs(deployment_cache_dir)
                # If the action is "create" or "update", the variables that are
                # passed are those expected in the target date. If the action is
                # "delete", the variables that are passed are those in the
                # current state.
                if action in ("create", "update"):
                    variables = self.package.target[key].variables
                else:
                    variables = self.package.current[key].variables
                # Call `prepare` to prepare input files to subprocesses, and
                # retrieve a list of commands to execute in subprocesses
                try:
                    module_config = self.package.get_module_config(key)
                    LOGGER.debug("%s Executing prepare", key)
                    commands = config.MODULES[key.module].prepare(
                        key=key,
                        command=config.CLI["command"],
                        action=action,
                        variables=variables,
                        module_config=module_config,
                        deployment_cache_dir=deployment_cache_dir,
                        engine_cache_dir=self.engines_cache_dir[
                            config.MODULES[key.module].engine
                        ],
                    )
                except:
                    section_that_failed = "prepare"
                    raise
                # Execute subprocesses
                for command in commands:
                    env = os.environ.copy()
                    env.update(command.env)
                    # Assume an IAM role if needed and add AWS temporary
                    # credentials to the environment variables
                    iam_role = module_config.get("AssumeRole")
                    if command.assume_role and not iam_role is None:
                        self._add_aws_temp_credentials(env, iam_role)
                    # Execute the subprocess
                    LOGGER.debug("%s Executing subprocess '%s'", key, command.name)
                    LOGGER.debug("%s Command: %s", key, " ".join(command.args))
                    LOGGER.debug("%s Cwd: %s", key, command.cwd)
                    try:
                        # pylint: disable=subprocess-popen-preexec-fn
                        # Need to catch SIGINT signals to exit gracefully
                        process = subprocess.Popen(
                            command.args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.PIPE,
                            env=env,
                            cwd=command.cwd,
                            shell=False,
                            start_new_session=True,
                        )
                        sent_sigint = False
                        sent_sigterm = False
                        while True:
                            # Check every second if a signal SIGINT or SIGTERM must
                            # be sent to be the subprocess. Wait for the subprocess
                            # to exit after sending the signal
                            try:
                                stdout, stderr = process.communicate(timeout=0.1)
                                # Write the subprocess logs
                                self._write_subprocess_logs(
                                    key,
                                    command.name,
                                    nb_attempts,
                                    "stdout.log",
                                    stdout.decode(),
                                )
                                self._write_subprocess_logs(
                                    key,
                                    command.name,
                                    nb_attempts,
                                    "stderr.log",
                                    stderr.decode(),
                                )
                                # Write stdout to a file if needed:
                                if not command.stdout_file is None:
                                    with open(command.stdout_file, "wb") as stream:
                                        stream.write(stdout)
                                # The subprocess must return an exit code of 0
                                if process.returncode != 0:
                                    section_that_failed = f"subprocess '{command.name}'"
                                    raise SubprocessError("Exit code is not 0")
                                # Interrupt the step if SIGINT or SIGTERM was
                                # pressed
                                if sent_sigint or sent_sigterm:
                                    section_that_failed = f"subprocess '{command.name}'"
                                    raise SubprocessError("Subprocess interrupted")
                                # Exit the loop if the subprocess has completed
                                break
                            except subprocess.TimeoutExpired:
                                if self.send_sigint.is_set() and not sent_sigint:
                                    process.send_signal(signal.SIGINT)
                                    sent_sigint = True
                                if self.send_sigterm.is_set() and not sent_sigterm:
                                    process.send_signal(signal.SIGTERM)
                                    sent_sigterm = True
                    except:
                        section_that_failed = f"subprocess '{command.name}'"
                        raise
                # Execute the post-process function to parse outputs from the
                # subprocesses
                try:
                    LOGGER.debug("%s Executing postprocess", key)
                    outcome = config.MODULES[key.module].postprocess(
                        key=key,
                        command=config.CLI["command"],
                        action=action,
                        module_config=module_config,
                        deployment_cache_dir=deployment_cache_dir,
                    )
                except:
                    section_that_failed = "postprocess"
                    raise
                # Complete the step if no exception was raised
                self.package.complete(
                    key,
                    made_changes=outcome.made_changes,
                    result=outcome.result,
                    detailed_results=outcome.detailed_results,
                    outputs=outcome.outputs,
                )
                LOGGER.info("%s Completed - %s", key, outcome.result)
                # Display the step execution time
                stop_time = datetime.now()
                elapsed_ms = (stop_time - start_time).total_seconds()
                LOGGER.debug("%s Execution time: %.3f seconds", key, elapsed_ms)
            # Catch any exceptions and mark the step as failed
            except Exception as err:
                LOGGER.error(
                    "%s Failed. See logs for details", key, exc_info=config.CLI["debug"]
                )
                self.package.fail(
                    key,
                    result="Failed",
                    detailed_results={
                        "FailedSection": section_that_failed,
                        "ErrorMessage": str(err),
                    },
                )
            # Delete the deployment cache for this deployment if needed
            finally:
                if self.delete_deployment_cache:
                    shutil.rmtree(deployment_cache_dir, ignore_errors=True)

    def _add_aws_temp_credentials(self, env: Dict[str, Any], iam_role: str) -> None:
        """Add AWS temporary credentials as environment variables to the dict
        `env`.

        Args:
            env (dict): Dictionary to which environment variables are added
            iam_role (str): ARN of the IAM role to assume
        """
        with self.lock_aws_temp_credentials:
            credentials = None
            # If AWS temporary credentials was created for this role less than
            # `CACHE_AWS_TEMP_CREDS` seconds ago, reuse the credentials
            if iam_role in self.aws_temp_credentials:
                created_at = self.aws_temp_credentials[iam_role]["CreatedAt"]
                created_since = (datetime.utcnow() - created_at).total_seconds()
                if created_since < config.CACHE_AWS_TEMP_CREDS:
                    credentials = self.aws_temp_credentials[iam_role]["Credentials"]
            # Otherwise, assume the role and add AWS temporary credentials to the
            # cache
            if credentials is None:
                credentials = utils.get_aws_temp_credentials(iam_role)
                self.aws_temp_credentials[iam_role] = {
                    "CreatedAt": datetime.utcnow(),
                    "Credentials": credentials,
                }
            # Add environment variables
            env["AWS_ACCESS_KEY_ID"] = credentials["AccessKeyId"]
            env["AWS_SECRET_ACCESS_KEY"] = credentials["SecretAccessKey"]
            env["AWS_SESSION_TOKEN"] = credentials["SessionToken"]

    def _write_subprocess_logs(
        self,
        key: ModuleAccountRegionKey,
        name_command: str,
        nb_attempts: int,
        filename: str,
        content: str,
    ) -> None:
        """Append subprocess logs to the local disk.

        Args:
            root_logs_dir: Root path of the folder that stores logs for this run
            key: Step key
            name_command: Name of the subprocess
            nb_attempts: Current attempt number
            filename: Name of the file to which logs must be appended
            content: Logs to append
        """
        # Create the folder for the log of this module deployment if it does not
        # exist
        log_path = os.path.join(
            self.root_logs_dir, key.module, key.account_id, key.region
        )
        if not os.path.isdir(log_path):
            os.makedirs(log_path)
        # Write the logs
        log_file = os.path.join(log_path, filename)
        with open(log_file, "a", encoding="utf-8") as stream:
            stream.write("################################\n")
            stream.write(f"# Subprocess '{name_command}'' - Attempt #{nb_attempts}\n")
            stream.write("################################\n")
            stream.write(content)
            stream.write("\n")

    # pylint: disable=unused-argument
    # The arguments `sig` and `frame` are mandatory but unused

    def _catch_sigint(self, sig, frame) -> None:
        """Catch SIGINT signals while workers are running. The first time
        CTRL+C is pressed, workers must stop after completing their current
        step. The second time, workers must be send a SIGINT signal to the
        running subprocess and wait for a graceful exit. The third time,
        workers must send a SIGTERM and wait subprocesses to exit. The fourth
        time, a KeyboardInterrupt exception is raised to interrupt this
        program.
        """

        self.nb_times_sigint += 1
        if self.nb_times_sigint == 1:
            self.stop_workers.set()
            LOGGER.info("Interrupted - Waiting for current deployments to complete")
        elif self.nb_times_sigint == 2:
            self.send_sigint.set()
            LOGGER.info("Interrupted - Sending SIGINT to subprocesses")
        elif self.nb_times_sigint == 3:
            self.send_sigterm.set()
            LOGGER.info("Interrupted - Sending SIGTERM to subprocesses")
        elif self.nb_times_sigint == 4:
            self.package.save()
            LOGGER.info("Interrupted - Forcing deployments to abort")
            raise KeyboardInterrupt
        else:
            raise KeyboardInterrupt

    @staticmethod
    def _stop_catching_sigint(sig, frame) -> None:
        """Stop catching SIGINT."""
        raise KeyboardInterrupt



"""Base class inherited by engine classes."""

import hashlib
import json
import logging
import os
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional

from aws_orga_deployer import config
from aws_orga_deployer.package.store import ModuleAccountRegionKey

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


@dataclass
class StepCommand:
    """Details about a subprocess to execute.

    Attributes:
        name: Friendly name for the subprocess
        args: List of subprocess arguments.
        cwd: Current working directory of the subprocess.
        assume_role: True if the IAM role must be assumed and AWS temporary
            credentials must be provided as environment variables to the
            subprocess.
        env: Additional environment variables to pass to the subprocess.
            Default to an empty dict (no additional environment variables).
        stdout_file: To save the content of the subprocess standard output
            `stdout` to a file, provide a file path.
    """

    name: str
    args: List[str]
    cwd: str
    assume_role: bool
    env: Dict[str, str] = field(default_factory=dict)
    stdout_file: Optional[str] = None


@dataclass
class StepOutcome:
    """Contains the results of the step execution.

    Attributes:
        made_changes: True if the step resulted in changes made or to be made.
        result: Summary of the result.
        detailed_results: Optional detailed results, such as the list of
            resources added, changed or deleted. Should only be valued for
            completed steps.
        outputs: Outputs returned when the step has completed. Should be valued
            only when the command is "apply" and if the action is not "destroy".
    """

    made_changes: bool
    result: str
    detailed_results: Optional[Dict[str, Any]] = None
    outputs: Optional[Dict[str, Any]] = None


class BaseEngine:
    """Base class inherited by engine classes.

    Attributes:
        name: Name of the module.
        engine: Name of the engine.
        module_dir: Path to the module directory.
        default_included_patterns: Default file patterns to include for
            computing the module hash of all modules for this engine, if no
            custom hash configuration is provided.
        default_excluded_patterns: Default file patterns to exclude for
            computing the module hash of all modules for this engine, if no
            custom hash configuration is provided.
        included_patterns: File patterns to include for this module (custom if
            a hash configuration file exists, or default).
        excluded_patterns: File patterns to exclude for this module (custom if
            a hash configuration file exists, or default).
        module_hash: Module hash.
    """

    # Default class attributes to override by engine classes
    engine: str = ""
    module_dir: str
    default_included_patterns: List[str] = ["*"]
    default_excluded_patterns: List[str] = []
    included_patterns: List[str]
    excluded_patterns: List[str]

    def __init__(self, name: str, path: str) -> None:
        """Create a BaseEngine instance.

        Args:
            name: Name of the module
            path: Path of module files
        """
        self.name = name
        self.module_dir = path
        self._get_filename_patterns_to_hash()
        self.module_hash = self._generate_module_hash()

    def _generate_module_hash(self) -> str:
        """Return the MD5 hash of the module."""
        return self._md5_update_from_dir(self.module_dir, hashlib.md5()).hexdigest()

    def _get_filename_patterns_to_hash(self) -> None:
        """Set the values of `self.included_patterns` and
        `self.excluded_patterns` that define the patterns of file names to
        use for computing the module hash.
        """
        self.included_patterns = self.default_included_patterns.copy()
        self.excluded_patterns = self.default_excluded_patterns.copy()
        try:
            # Load the hash configuration file if it exists in this module
            filename = os.path.join(self.module_dir, config.HASH_CONFIG_FILENAME)
            with open(filename, "rb") as stream:
                content = json.load(stream)
                assert isinstance(content, dict)
                if "Include" in content:
                    self.included_patterns = content["Include"]
                if "Exclude" in content:
                    self.excluded_patterns = content["Exclude"]
                for check in (self.included_patterns, self.excluded_patterns):
                    assert isinstance(check, list)
                    assert all(isinstance(item, str) for item in check)
            LOGGER.debug(
                "[%s] Found hash-config.json: Include=%s Exclude=%s",
                self.name,
                ",".join(self.included_patterns),
                ",".join(self.excluded_patterns),
            )
        except (FileNotFoundError, IOError):
            LOGGER.debug(
                "[%s] No hash-config.json file found. Using default: Include=%s"
                " Exclude=%s",
                self.name,
                ",".join(self.included_patterns),
                ",".join(self.excluded_patterns),
            )
        except (ValueError, AssertionError):
            LOGGER.debug(
                "[%s] hash-config.json is invalid. Using default: Include=%s"
                " Exclude=%s",
                self.name,
                ",".join(self.included_patterns),
                ",".join(self.excluded_patterns),
            )

    @staticmethod
    def _md5_update_from_file(filename: str, dir_hash: Any) -> Any:
        """Update the MD5 hash with a file."""
        assert Path(filename).is_file()
        with open(str(filename), "rb") as stream:
            for chunk in iter(lambda: stream.read(4096), b""):
                dir_hash.update(chunk)
        return dir_hash

    def _md5_update_from_dir(self, directory: str, dir_hash: Any) -> Any:
        """Update the MD5 hash by browsing files and folders recursively. Only
        the file names matching the patterns are used to compute the hash.
        """
        assert Path(directory).is_dir()
        # For each file or folder in this directory
        for path in sorted(Path(directory).iterdir()):
            # Browse subfolders recursively
            if path.is_dir():
                dir_hash = self._md5_update_from_dir(str(path), dir_hash)
            elif path.is_file():
                # Check that the filename matches one of the included patterns,
                # and not any of the excluded filename patterns. The hash
                # configuration file is also excluded
                is_included = False
                is_excluded = False
                for pattern in self.included_patterns:
                    expr = re.escape(pattern).replace("\\*", ".*").lower()
                    if re.match(expr, path.name.lower()):
                        is_included = True
                        break
                for pattern in self.excluded_patterns:
                    expr = re.escape(pattern).replace("\\*", ".*").lower()
                    if re.match(expr, path.name.lower()):
                        is_excluded = True
                        break
                if (
                    not is_included
                    or is_excluded
                    or path.name.endswith(config.HASH_CONFIG_FILENAME)
                ):
                    continue
                # Update the hash with the file name and content
                dir_hash.update(path.name.encode())
                dir_hash = BaseEngine._md5_update_from_file(str(path), dir_hash)
        return dir_hash

    def validate_module_config(self, module_config: Dict[str, Any]) -> None:
        """Validate the content of the module configuration.

        Args:
            module_config: Module configuration.

        Raises:
            AssertionError
        """
        # Check that AssumeRole is None or a string
        assume_role = module_config.get("AssumeRole")
        assert assume_role is None or isinstance(
            assume_role, str
        ), 'AssumeRole must be "null" or a string'
        # Check Retry parameters
        if "Retry" in module_config:
            if "MaxAttempts" in module_config["Retry"]:
                max_attempts = module_config["Retry"]["MaxAttempts"]
                assert isinstance(max_attempts, int), "MaxAttempts must be an integer"
                assert max_attempts > 0, "MaxAttempts must be larger than 0"
            if "DelayBeforeRetrying" in module_config["Retry"]:
                delay = module_config["Retry"]["DelayBeforeRetrying"]
                assert isinstance(delay, int), "DelayBeforeRetrying must be an integer"
                assert (
                    delay >= 0
                ), "DelayBeforeRetrying must be larger than or equal to 0"
        # Check EndpointUrls
        if "EndpointUrls" in module_config:
            assert isinstance(
                module_config["EndpointUrls"], dict
            ), "EndpointUrls must be a dict"

    def prepare(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        variables: Dict[str, Any],
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
        engine_cache_dir: str,
    ) -> List[StepCommand]:
        """Prepare inputs and return a list of commands to execute in
        subprocesses.

        Args:
            key: Step key to execute.
            command: CLI command requested ("preview" or "deploy").
            action: Action to be made ("create", "update" or "destroy").
            variables: Input variables.
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.
            engine_cache_dir: Path to the engine cache directory (common to all
                modules for this engine).

        Returns:
            List of commands to execute in subprocesses.
        """
        raise NotImplementedError

    def postprocess(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
    ) -> StepOutcome:
        """Parse the files that are generated by the subprocesses and return
        step outcomes.

        Args:
            key: Step key to execute.
            command: CLI command requested.
            action: Action to be made ("create", "update", "destroy").
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.

        Returns:
            Outcomes of the step.
        """
        raise NotImplementedError

    def write_wrapper_inputs(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        variables: Dict[str, Any],
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
        engine_cache_dir: str,
    ) -> None:
        """When using a Python wrapper script, write the inputs of the `prepare`
        function to a JSON file `input.json` that can be read by the wrapper.

        Args:
            key: Step key to execute.
            command: CLI command requested ("preview" or "deploy").
            action: Action to be made ("create", "update" or "destroy").
            variables: Input variables.
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.
            engine_cache_dir: Path to the engine cache directory (common to all
                modules for this engine).
        """
        # Prepare the content of the JSON file
        input_content: Dict[str, Any] = {
            "Deployment": key.to_dict(),
            "Command": command,
            "Action": action,
            "Variables": variables,
            "ModuleConfig": module_config,
            "ModulePath": self.module_dir,
            "DeploymentCacheDir": deployment_cache_dir,
            "EngineCacheDir": engine_cache_dir,
        }
        # Write the content to a file `input.json`
        input_file = os.path.join(deployment_cache_dir, "input.json")
        with open(input_file, "w", encoding="utf-8") as stream:
            json.dump(input_content, stream)

    def read_wrapper_outputs(self, deployment_cache_dir: str) -> StepOutcome:
        """When using a Python wrapper script, read the outputs of the wrapper
        script `output.json` and return a `StepOutcome` object.

        Args:
            deployment_cache_dir: Path to the module deployment cache directory.

        Returns:
            StepOutcome object.
        """
        output_file = os.path.join(deployment_cache_dir, "output.json")
        with open(output_file, "r", encoding="utf-8") as stream:
            content = json.load(stream)
            made_changes = content["MadeChanges"]
            result = content["Result"]
            detailed_results = content["DetailedResults"]
            outputs = content["Outputs"]
        return StepOutcome(made_changes, result, detailed_results, outputs)


def get_python_executable(module_config: Dict[str, Any]) -> str:
    """Return the path to the Python executable.

    Args:
        module_config: Module configuration.

    Returns:
        Path to the Python executable.
    """
    # Default value if no custom value is provided in the module configuration
    python_exec = "python3"
    if not module_config.get("PythonExecutable") is None:
        python_exec = module_config["PythonExecutable"]
    return python_exec



"""Main program module."""

import sys

from aws_orga_deployer import cli, config, execution, modules, utils
from aws_orga_deployer.orga import OrgaParser
from aws_orga_deployer.package import Package


def main() -> None:
    """Main function."""

    # pylint: disable=broad-exception-caught
    # We want to catch any exceptions that was uncatched before

    # Parse CLI arguments and retrieve a logger for this module
    logger = cli.init_cli()
    try:
        # Load modules and package configuration settings, and retrieve
        # information about AWS accounts and organizational units. The
        # `OrgaParser` must be instantiated after `Package` because it needs
        # settings from the package definition file
        modules.load_modules()
        package = Package()
        orga = OrgaParser()

        # If the requested command is "orga", export details about the AWS
        # accounts and organizational units
        if config.CLI["command"] == "orga":
            utils.write_output_json(
                orga.export(),
                "AWS account list and organization structure",
            )
            return

        # Retrieve deployed modules and determine the changes to be made. This
        # is not needed for the command "orga"
        package.full_init(orga)

        # If the requested command is "remove-orphans", remove the module
        # deployments to destroy that correspond to accounts that no longer
        # exist in the organization, or regions that are not longer enabled in
        # an account
        if config.CLI["command"] == "remove-orphans":
            dry_run = config.CLI["dry_run"]
            orphans_removed = package.remove_orphans(dry_run)
            utils.write_output_json(
                {"OrphanedDeployments": orphans_removed},
                "the list of orphaned module deployments",
            )
            if len(orphans_removed) > 0 and config.CLI["detailed_exitcode"]:
                sys.exit(2)
            sys.exit(0)

        # Determine changes to be made
        has_pending_changes = package.analyze_changes()

        # If the requested command is "list", export details about deployed
        # modules and changes to be made
        if config.CLI["command"] == "list":
            utils.write_output_json(
                package.export_changes(),
                "the list of deployed modules and changes to be made",
            )
            if has_pending_changes and config.CLI["detailed_exitcode"]:
                sys.exit(2)
            sys.exit(0)

        # If there are no changes to be made, the commands "preview",
        # "apply" or "update-hash" can be interrupted
        if not has_pending_changes:
            sys.exit(0)

        # Print information about the requested command
        if config.CLI["command"] == "preview":
            logger.info(
                '"preview" will determine which resources to add, update or delete if'
                " the pending deployments are applied"
            )
        elif config.CLI["command"] == "apply":
            logger.info(
                '"apply" will apply pending deployments, resulting in the creation,'
                " update or deletion of resources"
            )
        elif config.CLI["command"] == "update-hash":
            logger.info(
                '"update-hash" will update the value of the module hash for deployments'
                " to update"
            )

        # Ask confirmation of the deployment scope, unless the tool is used
        # in non-interactive mode
        if not config.CLI["non_interactive"]:
            val = input(
                'Enter "yes" to confirm the deployment scope (use the command "list"'
                " for details): "
            )
            if val != "yes":
                sys.exit(0)

        # Launch execution
        executor = execution.Executor(package)
        executor.run()

        # Display and export execution results
        made_changes, has_failed = package.analyze_results()
        utils.write_output_json(
            package.export_results(),
            "the result of the execution",
        )
        if has_failed:
            sys.exit(1)
        if made_changes and config.CLI["detailed_exitcode"]:
            sys.exit(2)
        sys.exit(0)

    except KeyboardInterrupt:
        logger.critical("Interrupted")
        sys.exit(1)
    except Exception as err:
        logger.critical(err, exc_info=config.CLI["debug"])
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Module with classes used to store information about module deployments."""

# COMPLETED
import logging
import time
from collections import UserDict
from copy import deepcopy
from threading import Thread
from typing import Any, Dict, List

from aws_orga_deployer import config, utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class ModuleAccountRegionKey:
    """Hashable class that can be used as key for stores and graphes related
    to module deployments.

    Attributes:
        module: Module name
        account_id: Account ID
        region: Region
    """

    module: str
    account_id: str
    region: str

    def __init__(self, *args: Any):
        """A dict or 3 values (module, account_id, region) can be provided."""
        if len(args) > 1:
            self.module = args[0]
            self.account_id = args[1]
            self.region = args[2]
        else:
            self.module = args[0]["Module"]
            self.account_id = args[0]["AccountId"]
            self.region = args[0]["Region"]

    def __eq__(self, another: object) -> bool:
        if not isinstance(another, ModuleAccountRegionKey):
            return False
        return (
            self.module == another.module
            and self.account_id == another.account_id
            and self.region == another.region
        )

    def __hash__(self):
        return hash((self.module, self.account_id, self.region))

    def __str__(self):
        return f"[{self.module},{self.account_id},{self.region}]"

    def to_dict(self) -> Dict[str, str]:
        """Export to a dictionary."""
        return {
            "Module": self.module,
            "AccountId": self.account_id,
            "Region": self.region,
        }


class TargetDeploymentDetails:
    """Store details about a module deployment in the target state.

    Attributes:
        variables: Variables used for this module deployment.
        var_from_outputs: Variables populated from the outputs of other module
            deployments.
        dependencies: List of module deployments that must be created before or
            deleted after this module deployment.
        module_hash: Current value of the module hash.
    """

    variables: Dict[str, Any]
    var_from_outputs: Dict[str, Any]
    dependencies: List[Dict[str, str]]
    module_hash: str

    def __init__(
        self,
        variables: Dict[str, Any],
        var_from_outputs: Dict[str, Any],
        dependencies: List[Dict[str, str]],
        module_hash: str,
    ) -> None:
        self.variables = variables
        self.var_from_outputs = var_from_outputs
        self.dependencies = dependencies
        self.module_hash = module_hash

    def __eq__(self, another: object) -> bool:
        if not isinstance(another, TargetDeploymentDetails):
            return False
        return (
            self.variables == another.variables
            and self.var_from_outputs == another.var_from_outputs
            and self.dependencies == another.dependencies
            and self.module_hash == another.module_hash
        )

    def to_dict(self) -> Dict[str, Any]:
        """Export to a dictionary."""
        return {
            "Variables": self.variables,
            "VariablesFromOutputs": self.var_from_outputs,
            "Dependencies": self.dependencies,
            "ModuleHash": self.module_hash,
        }


class CurrentDeploymentDetails:
    """Store details about a module deployment in the current state.

    Attributes:
        variables: Variables used for this module deployment.
        var_from_outputs: Variables populated from the outputs of other module
            deployments.
        dependencies: List of module deployments that must be created before or
            deleted after this module deployment.
        module_hash: Value of the module hash when the deployment was made.
        outputs: Outputs returned by this module deployment.
        last_changed_time: Last time this module deployment was changed.
    """

    variables: Dict[str, Any]
    var_from_outputs: Dict[str, Any]
    dependencies: List[Dict[str, str]]
    module_hash: str
    outputs: Dict[str, Any]
    last_changed_time: str

    def __init__(self, *args: Any, **kwargs: Any):
        """A dict generated with `to_dict` or 4 values can be provided."""
        if len(args) > 0 and isinstance(args[0], dict):
            self.variables = args[0]["Variables"]
            self.var_from_outputs = args[0]["VariablesFromOutputs"]
            self.dependencies = args[0]["Dependencies"]
            self.module_hash = args[0]["ModuleHash"]
            self.outputs = args[0]["Outputs"]
            self.last_changed_time = args[0]["LastChangedTime"]
        else:
            self.variables = args[0] if len(args) > 0 else kwargs["variables"]
            self.var_from_outputs = (
                args[1] if len(args) > 1 else kwargs["var_from_outputs"]
            )
            self.dependencies = args[2] if len(args) > 2 else kwargs["dependencies"]
            self.module_hash = args[3] if len(args) > 3 else kwargs["module_hash"]
            self.outputs = args[4] if len(args) > 4 else kwargs["outputs"]
            self.last_changed_time = (
                args[5] if len(args) > 5 else kwargs["last_changed_time"]
            )

    def __eq__(self, another: object) -> bool:
        if not isinstance(another, CurrentDeploymentDetails):
            return False
        return (
            self.variables == another.variables
            and self.var_from_outputs == another.var_from_outputs
            and self.dependencies == another.dependencies
            and self.outputs == another.outputs
            and self.module_hash == another.module_hash
            and self.last_changed_time == another.last_changed_time
        )

    def to_dict(self) -> Dict[str, Any]:
        """Export to a dictionary."""
        return {
            "Variables": self.variables,
            "VariablesFromOutputs": self.var_from_outputs,
            "Dependencies": self.dependencies,
            "ModuleHash": self.module_hash,
            "Outputs": self.outputs,
            "LastChangedTime": self.last_changed_time,
        }


class CurrentStateStore(UserDict):
    """Class that loads and saves the content of a dictionay from S3. The
    keys must be of type ModuleAccountRegionKey and the values must be of type
    CurrentDeploymentDetails.

    Attributes:
        period: Frequency at which the package state is saved to S3. If the
            period is 0, the periodic saving is disabled.
    """

    period: int

    def __init__(self, period: int = 0):
        def func() -> None:
            # pylint: disable = bare-except
            # We need to catch any exceptions so that this thread never stops

            while True:
                try:
                    time.sleep(period)
                    if self._must_stop is True:
                        return
                    self.save()
                except:
                    LOGGER.exception(
                        "Failed to save the package state to S3",
                        exc_info=config.CLI["debug"],
                    )

        # Load and deserialize the content from S3 if the package state exists.
        # If the file does not exist in S3, start with a blank dict
        dict_content = {}
        try:
            serialized = utils.load_json_from_s3(config.STATE_FILENAME)
            current_deployments = serialized["Deployments"]
            for item in current_deployments:
                key = ModuleAccountRegionKey(item["Deployment"])
                value = CurrentDeploymentDetails(item["CurrentState"])
                dict_content[key] = value
        except utils.FileNotExists:
            pass
        super().__init__(dict_content)
        self._must_stop: bool = False
        self._data_copy: Dict = deepcopy(self.data)
        # Initialize a thread that saves the state every `period` seconds if it
        # has changed since the last time it was saved to S3, unless disabled.
        if period > 0:
            Thread(target=func, daemon=True, name="CurrentStateStore").start()

    def save(self) -> bool:
        """Save the content to S3 if it has changed since the last time it was
        saved.

        Returns:
            True if the package state was saved to S3, False if no change was
                found.
        """
        if self.data == self._data_copy:
            return False
        LOGGER.debug("Saving the package state to S3")
        # Serialize the package state into a JSON document that contains a list
        # of current deployments, and each current deployment is a dict with:
        #
        # {
        #   `Deployment`: `ModuleAccountRegionKey.to_dict()`
        #   `CurrentState`: `CurrentDeploymentDetails.to_dict()`
        # }
        current_deployments = []
        for key, value in self.data.items():
            item = {"Deployment": key.to_dict(), "CurrentState": value.to_dict()}
            current_deployments.append(item)
        serialized = {"Deployments": current_deployments}
        # Write the file to S3, and store a copy of the content saved to S3
        # so that we can evaluate if the content has changed since the last
        # save to S3.
        utils.write_dict_to_s3(serialized, config.STATE_FILENAME)
        self._data_copy = deepcopy(self.data)
        return True

    def stop(self) -> None:
        """Force the daemon thread that saves the current stat at regular
        intervals to stop.
        """
        self._must_stop = True



"""Engine for Terraform modules."""

import json
import shutil
from os import path
from typing import Any, Dict, List

from aws_orga_deployer import config, utils
from aws_orga_deployer.engines import base
from aws_orga_deployer.package.store import ModuleAccountRegionKey


class Engine(base.BaseEngine):
    """Engine for Terraform modules."""

    engine = "terraform"
    default_included_patterns = ["*.tf"]
    default_excluded_patterns = []

    def validate_module_config(self, module_config: Dict[str, Any]) -> None:
        """Validate the content of the module configuration.

        Args:
            module_config: Module configuration.

        Raises:
            AssertionError
        """
        super().validate_module_config(module_config)
        if "TerraformExecutable" in module_config:
            assert isinstance(
                module_config["TerraformExecutable"], str
            ), "TerraformExecutable must be a string"

    def prepare(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        variables: Dict[str, Any],
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
        engine_cache_dir: str,
    ) -> List[base.StepCommand]:
        """Prepare Terraform files in the deployment cache directory and return
        a list of Terraform commands to execute in subprocesses.

        Args:
            key: Step key to execute.
            command: CLI command requested ("preview" or "deploy").
            action: Action to be made ("create", "update" or "destroy").
            variables: Input variables.
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.
            engine_cache_dir: Path to the engine cache directory (common to all
                modules for this engine).

        Returns:
            List of commands to execute in subprocesses.
        """
        # If the action is "create" or "update", copy the Terraform templates
        # that exist in the module directory to the deployment cache directory,
        # and prepare a file `terraform.tfvars.json`. If the action is "delete",
        # the directory doesn't contain any Terraform which means that the
        # target state must not contain any resources.
        if action in ("create", "update"):
            # The existing directory must not exist, or `copytree` fails with
            # Python 3.7
            shutil.rmtree(deployment_cache_dir, ignore_errors=True)
            shutil.copytree(self.module_dir, deployment_cache_dir)
            # Create the file `terraform.tfvars.json`
            var_filename = path.join(deployment_cache_dir, "terraform.tfvars.json")
            with open(var_filename, "w", encoding="utf-8") as stream:
                json.dump(variables, stream, indent=4)
        # Create a file `aws_orga_deployer.tf` that contains the configuration
        # of the AWS provider and of the S3 backend
        tf_filename = path.join(deployment_cache_dir, "aws_orga_deployer.tf")
        with open(tf_filename, "w", encoding="utf-8") as stream:
            stream.write('provider "aws" {\n')
            stream.write(f'  region = "{key.region}"\n')
            # If an IAM role must be assumed in the target AWS account, it must
            # be assumed by the provider, so that the S3 backend has permissions
            # in the current execution account to write in the package state
            # bucket.
            if not module_config.get("AssumeRole") is None:
                stream.write("  assume_role {\n")
                stream.write(f'    role_arn = "{module_config["AssumeRole"]}"\n')
                stream.write('    session_name = "aws-orga-deployer"\n')
                stream.write("  }\n")
            # Add custom endpoints for the AWS provider if needed
            if "EndpointUrls" in module_config:
                stream.write("  endpoints {\n")
                for service, url in module_config["EndpointUrls"].items():
                    stream.write(f'    {service} = "{url}"\n')
                stream.write("  }\n")
            stream.write("}\n")
            stream.write("terraform {\n")
            stream.write('  backend "s3" {\n')
            stream.write(f'    bucket = "{config.PACKAGE["S3Bucket"]}"\n')
            stream.write(f'    region = "{config.PACKAGE["S3Region"]}"\n')
            s3_key = utils.get_s3_key(
                f"terraform/{key.module}/{key.account_id}/{key.region}/terraform.tfstate"
            )
            stream.write(f'    key = "{s3_key}"\n')
            # Add custom endpoints for S3, STS and IAM for the S3 backend
            if "EndpointUrls" in module_config:
                for service, url in module_config["EndpointUrls"].items():
                    if service == "s3":
                        stream.write(f'    endpoint = "{url}"\n')
                        stream.write("    force_path_style = true\n")

                    elif service == "sts":
                        stream.write(f'    sts_endpoint = "{url}"\n')
                    elif service == "iam":
                        stream.write(f'    iam_endpoint = "{url}"\n')
            stream.write("  }\n")
            stream.write("}\n")
        # Get the path to the Terraform executable
        terraform_exec = "terraform"
        if not module_config.get("TerraformExecutable") is None:
            terraform_exec = module_config["TerraformExecutable"]
        # Set arguments and environment variables that are common to all
        # Terraform commands
        common_args = ["-no-color"]
        common_env = {
            "TF_PLUGIN_CACHE_DIR": engine_cache_dir,
            "TF_PLUGIN_CACHE_MAY_BREAK_DEPENDENCY_LOCK_FILE": "true",
        }
        # `terraform init` is needed whether the command or the action
        commands = [
            base.StepCommand(
                name="init",
                args=[terraform_exec, "init", *common_args],
                cwd=deployment_cache_dir,
                assume_role=False,
                env=common_env,
            )
        ]
        # The following commands are needed for both "preview" and "apply"
        # commands
        commands += [
            base.StepCommand(
                name="plan",
                args=[terraform_exec, "plan", "-out=tfplan", *common_args],
                cwd=deployment_cache_dir,
                assume_role=False,
                env=common_env,
            ),
            # The output of the following command must be stored in a file so
            # that the `postprocess` is able to identify the resources to
            # add, change or delete
            base.StepCommand(
                name="get plan in JSON",
                args=[terraform_exec, "show", "-json", "tfplan", *common_args],
                cwd=deployment_cache_dir,
                assume_role=False,
                env=common_env,
                stdout_file=path.join(deployment_cache_dir, "plan.json"),
            ),
        ]
        # The following commands are only run if command is "apply" to apply
        # the plan and get outputs
        if command == "apply":
            commands += [
                base.StepCommand(
                    name="apply plan",
                    args=[
                        terraform_exec,
                        "apply",
                        "-auto-approve",
                        *common_args,
                        "tfplan",
                    ],
                    cwd=deployment_cache_dir,
                    assume_role=False,
                    env=common_env,
                )
            ]
            # If the action is not "destroy", get the outputs from the Terraform
            # state. The output must be stored in a file, so that `postprocess`
            # can read the outputs.
            if action in ("create", "update"):
                commands += [
                    base.StepCommand(
                        name="get outputs",
                        args=[terraform_exec, "output", "-json", *common_args],
                        cwd=deployment_cache_dir,
                        assume_role=False,
                        env=common_env,
                        stdout_file=path.join(deployment_cache_dir, "output.json"),
                    ),
                ]
        return commands

    def postprocess(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
    ) -> base.StepOutcome:
        """Reads the file `output.json` created by the "wrapper" function and
        returns the outputs.

        Args:
            key: Step key to execute.
            command: CLI command requested.
            action: Action to be made ("create", "update", "destroy").
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.

        Returns:
            Outcome of the step.
        """
        # Identify the resources to add, change or delete from the Terraform
        # plan
        res_add = []
        res_change = []
        res_delete = []
        plan_file = path.join(deployment_cache_dir, "plan.json")
        with open(plan_file, "r", encoding="utf-8") as stream:
            content = json.load(stream)
            for change in content.get("resource_changes", []):
                actions = change["change"]["actions"]
                if actions == ["create"]:
                    res_add.append(change["address"])
                elif actions == ["delete"]:
                    res_delete.append(change["address"])
                # Actions ["delete", "create"] or ["create", "delete"]
                # correspond to "update" because the resource is deleted and
                # recreated
                elif actions in (
                    ["update"],
                    ["delete", "create"],
                    ["create", "delete"],
                ):
                    res_change.append(change["address"])
        # If the command is "preview", the outputs must be named "to add", "to
        # change" or to "to delete" and no outputs is needed
        if command == "preview":
            return base.StepOutcome(
                made_changes=len(res_add + res_change + res_delete) > 0,
                result=(
                    f"{len(res_add)} resources to add, "
                    f"{len(res_change)} to change, "
                    f"{len(res_delete)} to delete"
                ),
                detailed_results={
                    "ResourcesToAdd": res_add,
                    "ResourcesToChange": res_change,
                    "ResourcesToDelete": res_delete,
                },
            )
        # If the command is "apply", the outcomes must be named "added",
        # "changed" or "deleted" and outputs must be provided if action is
        # "create" or "update"
        if action in ("create", "update"):
            output_file = path.join(deployment_cache_dir, "output.json")
            with open(output_file, "r", encoding="utf-8") as stream:
                content = json.load(stream)
                outputs = {key: value["value"] for key, value in content.items()}
        else:
            outputs = None
        return base.StepOutcome(
            made_changes=len(res_add + res_change + res_delete) > 0,
            result=(
                f"{len(res_add)} resources added, "
                f"{len(res_change)} changed, "
                f"{len(res_delete)} deleted"
            ),
            detailed_results={
                "ResourcesAdded": res_add,
                "ResourcesChanged": res_change,
                "ResourcesDeleted": res_delete,
            },
            outputs=outputs,
        )



"""Wrapper of CloudFormation modules that is executed in a subprocess."""

import random
import string
import time
from os import path
from typing import Any, Dict, List

import boto3
import botocore

from aws_orga_deployer.engines.wrappers import base

# CloudFormation capabilities passed to API requests
CFN_CAPABILITIES: List[str] = [
    "CAPABILITY_IAM",
    "CAPABILITY_NAMED_IAM",
    "CAPABILITY_AUTO_EXPAND",
]


def main() -> None:
    """Main function."""

    def check_stack_exists() -> bool:
        """Return True to the stack exists."""
        try:
            _stack = client.describe_stacks(StackName=stack_name)["Stacks"][0]
            _status = _stack["StackStatus"]
            # The stack has the status REVIEW_IN_PROGRESS when change sets
            # were created but the stack has not yet been created
            if _status == "REVIEW_IN_PROGRESS":
                return False
            return True
        except botocore.exceptions.ClientError as err:
            if "does not exist" in str(err):
                return False
            raise err

    def get_additional_boto3_params() -> Dict[str, Any]:
        """Retrieve and returns the list of additional parameters that must be
        passed to boto3 API requests.
        """
        if "AdditionalBoto3Parameters" in inputs.module_config:
            return inputs.module_config["AdditionalBoto3Parameters"]
        return {}

    def get_template_body() -> str:
        """Load the template whose filename is given in `TemplateFilename` and
        returns its content.
        """
        filename = inputs.module_config["TemplateFilename"]
        try:
            filepath = path.join(inputs.module_dir, filename)
            with open(filepath, "r", encoding="utf-8") as stream:
                return stream.read()
        except OSError as err:
            raise ValueError(f"Cannot read the template at {filepath}") from err

    def get_stack_status() -> str:
        """Returns the stack status."""
        _stack = client.describe_stacks(StackName=stack_name)["Stacks"][0]
        return _stack["StackStatus"]

    def get_stack_outputs() -> Dict[str, str]:
        """Returns the stack outputs."""
        time.sleep(2)
        _stack = client.describe_stacks(StackName=stack_name)["Stacks"][0]
        if not "Outputs" in _stack:
            return {}
        return {
            output["OutputKey"]: output["OutputValue"] for output in _stack["Outputs"]
        }

    def get_stack_resources() -> List[str]:
        """Returns the list of stack resources."""
        result = []
        paginator = client.get_paginator("list_stack_resources")
        pages = paginator.paginate(StackName=stack_name)
        for page in pages:
            for resource in page["StackResourceSummaries"]:
                result.append(resource["LogicalResourceId"])
        return result

    def prepare_stack_parameters() -> List[Dict[str, Any]]:
        """Generate and return a list of stack parameters from the input
        variables.
        """
        return [
            {
                "ParameterKey": name,
                "ParameterValue": ",".join(value) if isinstance(value, list) else value,
                "UsePreviousValue": False,
            }
            for name, value in inputs.variables.items()
        ]

    def delete_change_set() -> None:
        """Try to delete the change set, but ignore any exceptions raised by
        boto3.
        """
        try:
            client.delete_change_set(ChangeSetName=change_set_id)
        except botocore.exceptions.ClientError:
            pass

    # Read the inputs from the JSON file `input.json` and create a boto3 client
    # for the module region
    inputs = base.read_wrapper_inputs()
    # Create a boto3 client for CloudFormation with custom endpoint if needed
    endpoint_config = {}
    if "EndpointUrls" in inputs.module_config:
        if "cloudformation" in inputs.module_config["EndpointUrls"]:
            endpoint_url = inputs.module_config["EndpointUrls"]["cloudformation"]
            endpoint_config = {"endpoint_url": endpoint_url}
    client = boto3.client(
        "cloudformation", region_name=inputs.region, **endpoint_config
    )
    # Retrieve the stack name
    stack_name = inputs.module_config["StackName"]
    # Check if the stack already exists
    stack_exists = check_stack_exists()
    if stack_exists is True:
        print("The stack {stack_name} exists")
    else:
        print("The stack {stack_name} does not exist")
    #######################################
    # If the action is "create" or "update"
    if inputs.action in ("create", "update"):
        # Create a change set to identify the list of resources to add,
        # change or delete. The name of the change set is the name of the
        # stack, prepended by 5 random letters
        random_suffix = "".join(random.choices(string.ascii_lowercase, k=5))
        change_set_name = stack_name + "-" + random_suffix
        change_set_id = client.create_change_set(
            StackName=stack_name,
            TemplateBody=get_template_body(),
            UsePreviousTemplate=False,
            Parameters=prepare_stack_parameters(),
            Capabilities=CFN_CAPABILITIES,
            ChangeSetName=change_set_name,
            Description="Change set for AWS Orga Deployer",
            ChangeSetType="UPDATE" if stack_exists else "CREATE",
            IncludeNestedStacks=False,
            OnStackFailure="ROLLBACK" if stack_exists else "DELETE",
            **get_additional_boto3_params(),
        )["Id"]
        print(f"Creating the change set {change_set_name}")
        # Wait until the change set completes
        while True:
            time.sleep(3)
            change_set = client.describe_change_set(ChangeSetName=change_set_id)
            status = change_set["Status"]
            # Stop waiting if the change set has completed
            if status.endswith("COMPLETE"):
                print("The creation of the change set has completed")
                break
            # Wait if the status is still pending or in progress
            if status.endswith("IN_PROGRESS") or status.endswith("PENDING"):
                print("Waiting for the change set creation to complete")
                continue
            # For any other status
            print("Deleting the change set")
            delete_change_set()
            # Retrieve the existing outputs if the stack already exists,
            # because outputs must be provided if command is "apply"
            if stack_exists:
                existing_outputs = get_stack_outputs()
            else:
                existing_outputs = None
            # If the change set fails because there are no changes to be
            # made, no need to continue
            reason = change_set.get("StatusReason")
            if "information didn't contain changes" in reason:
                print("No changes to be made")
                base.write_wrapper_outputs(
                    made_changes=False,
                    result="No changes to be made",
                    detailed_results=None,
                    outputs=existing_outputs if inputs.command == "apply" else None,
                )
                return
            # Raise an error if the change set failed for another reason
            print(f"The creation of the change set failed: {status} {reason}")
            raise RuntimeError("Failed to evaluate the changes to be made")
        # Identify the resources to add, change or delete
        res_add = []
        res_change = []
        res_delete = []
        for change in change_set["Changes"]:
            if change["Type"] == "Resource":
                rsc_change = change["ResourceChange"]
                if rsc_change["Action"] == "Add":
                    res_add.append(rsc_change["LogicalResourceId"])
                elif rsc_change["Action"] == "Modify":
                    res_change.append(rsc_change["LogicalResourceId"])
                elif rsc_change["Action"] == "Remove":
                    res_delete.append(rsc_change["LogicalResourceId"])
        # If the command is preview, the outcomes must be named "to add",
        # "to change", "to delete" and there are no outputs to return
        if inputs.command == "preview":
            # Write the outputs
            base.write_wrapper_outputs(
                made_changes=True,
                result=(
                    f"{len(res_add)} resources to add, "
                    f"{len(res_change)} to change, "
                    f"{len(res_delete)} to delete"
                ),
                detailed_results={
                    "ResourcesToAdd": res_add,
                    "ResourcesToChange": res_change,
                    "ResourcesToDelete": res_delete,
                },
            )
        # If the command is preview, the outcomes must be named "added",
        # "changed", "deleted" and outputs must be returned
        elif inputs.command == "apply":
            # Execute the change set
            print("Executing the change set")
            client.execute_change_set(ChangeSetName=change_set_id)
            # Wait until the stack updates completes
            while True:
                time.sleep(3)
                status = get_stack_status()
                if status.endswith("COMPLETE") and not "ROLLBACK" in status:
                    print("The execution of the change set has completed")
                    break
                # Wait if the status is still pending or in progress
                if status.endswith("IN_PROGRESS"):
                    print("Waiting for the change set execution to complete")
                    continue
                # If the status is unknown, raise an error
                delete_change_set()
                print(f"The execution of the change set failed: {status}")
                raise RuntimeError("Failed to apply the changes to be made")
            # Write the outputs assuming that the changes made are those
            # identified in the change set
            base.write_wrapper_outputs(
                made_changes=True,
                result=(
                    f"{len(res_add)} resources added, "
                    f"{len(res_change)} changed, "
                    f"{len(res_delete)} deleted"
                ),
                detailed_results={
                    "ResourcesAdded": res_add,
                    "ResourcesChanged": res_change,
                    "ResourcesDeleted": res_delete,
                },
                outputs=get_stack_outputs(),
            )
        # Delete the change set and exit
        print("Deleting the change set")
        delete_change_set()
    #######################################
    # If the action is "destroy"
    elif inputs.action == "destroy":
        # Retrieve the list of existing resources if the stack exists
        if stack_exists:
            res_delete = get_stack_resources()
        else:
            res_delete = []
        # If the command is "preview", return the list of existing resources
        # as the list of resources to delete
        if inputs.command == "preview":
            base.write_wrapper_outputs(
                made_changes=len(res_delete) > 0,
                result=f"{len(res_delete)} resources to delete",
                detailed_results={"ResourcesToDelete": res_delete},
            )
        # If the command is "apply", delete the stack and return the list of
        # existing resources as the list of deleted resources
        elif inputs.command == "apply":
            # Delete the stack
            print("Deleting the stack")
            client.delete_stack(StackName=stack_name)
            # Wait for the stack deletion to complete
            while True:
                time.sleep(3)
                exists = check_stack_exists()
                if not exists:
                    break
                status = get_stack_status()
                if status.endswith("COMPLETE"):
                    print("The deletion of the stack has completed")
                    break
                # Wait if the status is still pending or in progress
                if status.endswith("IN_PROGRESS"):
                    print("Waiting for the stack deletion to complete")
                    continue
                # If the status is unknown, raise an error
                print(f"The deletion of the stack failed: {status}")
                raise RuntimeError("Failed to delete the stack")
            # Return the list of resources that existed before the stack
            # was deleted
            base.write_wrapper_outputs(
                made_changes=len(res_delete) > 0,
                result=f"{len(res_delete)} resources deleted",
                detailed_results={"ResourcesDeleted": res_delete},
            )


if __name__ == "__main__":
    main()



"""Test the module `aws_orga_deployer.package.__init__`."""

import time
import unittest
from unittest.mock import patch

from moto import mock_organizations, mock_s3, mock_sts

from aws_orga_deployer import modules
from aws_orga_deployer.orga import OrgaParser
from aws_orga_deployer.package import Package, PackageError, graph, store
from tests import mock
from tests.mock import mock_others
from tests.utils import update_cli_filters


class TestPackage(unittest.TestCase):
    """Test the class Package. At the end of each test, make sure to stop the
    auto-save of the current state store for each Package instance, or the
    daemon thread may continue to write to S3 after the test has completed,
    creating an inconsistent state.
    """

    @mock_sts
    @mock_others
    def setUp(self):
        """Create mock resources and an OrgaParser object that must exist for
        all tests.
        """
        mock.mock_cli_arguments(package_filename="package1.yaml")
        # Enable mocking for S3 and Organizations in all tests without having
        # to add declarators to each test function
        self.mock_s3 = mock_s3()
        self.mock_s3.start()
        mock.create_fake_bucket()
        # Load modules
        modules.load_modules()
        # Enable persistent mocking for Organizations
        self.mock_organizations = mock_organizations()
        self.mock_organizations.start()
        # Create a fake AWS organization
        self.fake_orga = mock.FakeOrgaParser()
        # Create an OrgaParser object that queries the fake organization.
        # A Package object must be created first to load package configuration
        # settings, but it is unused after
        Package()
        self.orga = OrgaParser()

    def tearDown(self):
        """Disable mocking for S3 and Organizations."""
        self.mock_s3.stop()
        self.mock_organizations.stop()

    def test_complete(self):
        """Complete a series of tests reflecting standard usage."""
        package1 = Package()
        package1.full_init(self.orga)
        # Check the number of module deployments in the target state. There
        # should be "python1" modules in all accounts in us-east-1 only, and
        # "terraform1" in all accounts except master in all enabled regions
        nb_expected = len(self.orga.accounts) + (
            len(mock.ENABLED_REGIONS) * (len(self.orga.accounts) - 1)
        )
        self.assertEqual(len(package1.target), nb_expected)
        # Check the result of `analyze_changes`. It should return that there
        # are pending changes
        has_pending_changes = package1.analyze_changes()
        self.assertTrue(has_pending_changes)
        # Check the result of `export_changes`. It should contain the same
        # number of items in `result["PendingChanges"]["Create"]`
        export = package1.export_changes()
        self.assertEqual(len(export["PendingChanges"]["Create"]), nb_expected)
        # Check the content of one specific item in the target state
        key = store.ModuleAccountRegionKey(
            "terraform1", self.fake_orga.account_a_id, mock.ENABLED_REGIONS[0]
        )
        self.assertIn(key, package1.target)
        self.assertIn("var1", package1.target[key].variables)
        self.assertIn("terraform1", package1.target[key].variables["varGlobal"])
        # Check the module configuration and that the value of `AssumeRole`
        # matches the account ID
        module_config = package1.get_module_config(key)
        self.assertEqual(
            module_config["AssumeRole"],
            f"arn:aws:iam::{self.fake_orga.account_a_id}:role/Admin",
        )
        # Retrieve the next step to process and check that it matches a step
        # for "python1" module in the us-east-1 region
        key, action, nb_attempts, max_attempts = package1.next()
        self.assertEqual(key.module, "python1")
        self.assertEqual(key.region, "us-east-1")
        self.assertEqual(action, "create")
        self.assertEqual(nb_attempts, 1)
        self.assertEqual(max_attempts, 2)
        # Complete the step
        package1.complete(
            key,
            made_changes=True,
            result="Summary",
            detailed_results={"keyDetail": "valueDetail"},
            outputs={"outputPython1": "valueOutput"},
        )
        # Check that the current state contains one item for the step that was
        # just completed and check its value
        self.assertIn(key, package1.current)
        self.assertEqual(
            package1.current[key].outputs["outputPython1"],
            "valueOutput",
        )
        # Check the results export. There should be one completed create step
        export = package1.export_results()
        self.assertEqual(len(export["Completed"]["Create"]), 1)
        self.assertEqual(export["Completed"]["Create"][0]["ResultedInChanges"], True)
        # Reload the package and check that the current state is loaded from S3
        package1.save()
        package2 = Package()
        package2.full_init(self.orga)
        self.assertIn(key, package2.current)

    def test_update(self):
        """Test the class Package with deployments to update."""
        package1 = Package()
        package1.full_init(self.orga)
        # Complete a deployment and change the value of the variables in the
        # current state to simulate that an update is needed
        key, _, _, _ = package1.next()
        package1.complete(key, made_changes=True, result="Summary")
        package1.current[key].variables["varPython1"] = "old_value"
        package1.save()
        # Reload the package and check that there is one pending update to
        # make
        package2 = Package()
        package2.full_init(self.orga)
        export = package2.export_changes()
        self.assertEqual(len(export["PendingChanges"]["Update"]), 1)
        # Complete all deployments
        try:
            while True:
                key, _, _, _ = package2.next()
                package2.complete(key, made_changes=True, result="Summary")
        except graph.NoMorePendingStep:
            pass
        # Reload the package and check that there are pending changes
        package2.save()
        package3 = Package()
        package3.full_init(self.orga)
        self.assertFalse(package3.analyze_changes())

    # Change the command to preview
    def test_update_preview_mode(self):
        """Test the class Package with deployments to update."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"command": "preview"}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Complete all deployments
            try:
                while True:
                    key, _, _, _ = package1.next()
                    package1.complete(key, made_changes=True, result="Summary")
            except graph.NoMorePendingStep:
                pass
            # Reload the package and check that there is are still pending
            # deployments because it was preview mode
            package1.save()
            package2 = Package()
            package2.full_init(self.orga)
            self.assertTrue(package2.analyze_changes())

    def test_update_hash(self):
        """Test the `update_hash` function."""
        package1 = Package()
        package1.full_init(self.orga)
        # Retrieve the next step and change the value of the module hash in the
        # current state to simulate that an update is needed because the value
        # of the module hash differs
        key, _, _, _ = package1.next()
        package1.complete(key, made_changes=True, result="Summary")
        package1.current[key].module_hash = "fake_hash"
        # Retrieve another step and change the value of the variables
        key, _, _, _ = package1.next()
        package1.complete(key, made_changes=True, result="Summary")
        package1.current[key].variables = {"var2": "value2"}
        # Reload the package and execute the `update_hash` function for all
        # steps, until there are no more steps to process
        package1.save()
        package2 = Package()
        package2.full_init(self.orga)
        try:
            while True:
                key, _, _, _ = package2.next()
                package2.update_hash(key)
        except graph.NoMorePendingStep:
            pass
        # Check that there two updates in the results (the other steps are
        # pending creation)
        export = package2.export_results()
        self.assertEqual(len(export["Completed"]["Update"]), 2)
        # Check that one has resulted in changes (module hash changed) and the
        # other no (change in variables)
        made_changes = set(
            item["ResultedInChanges"] for item in export["Completed"]["Update"]
        )
        self.assertEqual(made_changes, set([True, False]))
        # Reload the package and there should be one step with no changes to
        # be made (module hash changed) and one with update (variables updated)
        package2.save()
        package3 = Package()
        package3.full_init(self.orga)
        export = package3.export_changes()
        self.assertEqual(len(export["NoChanges"]), 1)
        self.assertEqual(len(export["PendingChanges"]["Update"]), 1)

    def test_destroy(self):
        """Test the class Package with deployments to destroy."""
        package1 = Package()
        package1.full_init(self.orga)
        # Retrieve a step to process and complete the deployment
        key, _, _, _ = package1.next()
        package1.complete(key, made_changes=True, result="Summary")
        # Reload the package and remove the deployments in the target state to
        # simulate that a resource exists in the current state, but not in the
        # target state and it must be destroyed
        package1.save()
        package2 = Package()
        for module_block in package2.package["Modules"].values():
            module_block["Deployments"] = []
        package2.full_init(self.orga)
        # Check that there are deployments to destroy
        result = package2.export_changes()
        self.assertEqual(len(result["PendingChanges"]["Destroy"]), 1)
        # Complete the deployment
        key, _, _, _ = package2.next()
        package2.complete(key, made_changes=True, result="Summary")
        # Reload the package and check that there are no deployments to destroy
        package2.save()
        package3 = Package()
        package3.full_init(self.orga)
        result = package3.export_changes()
        self.assertNotIn("Delete", result["PendingChanges"])

    def test_nonexistent_module(self):
        """Test the class Package with deployments in the current state for
        which there is no module block in the package definition file.
        """
        package1 = Package()
        package1.full_init(self.orga)
        # Retrieve a step to process and complete the deployment
        key, _, _, _ = package1.next()
        package1.complete(key, made_changes=True, result="Summary")
        package1.save()
        # Reload the package and remove all the module blocks. It should raise
        # a PackageError exception
        package2 = Package()
        package2.package["Modules"] = {}
        with self.assertRaises(PackageError):
            package2.full_init(self.orga)

    # Add a filter to include only the "us-east-1" region
    def test_cli_filters(self):
        """Test the class Package with CLI filters."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"include_regions": ["us-east-1"]}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Check the number of pending changes that are not skipped. There
            # should be "python1" in all accounts in us-east-1 only, and
            # "terraform1" in all accounts except master in us-east-1 only
            self.assertTrue(package1.analyze_changes())
            nb_pending = len(self.orga.accounts) * 2 - 1
            export = package1.export_changes()
            self.assertEqual(len(export["PendingChanges"]["Create"]), nb_pending)
            # Check that the number of steps returned matches this number
            nb_steps_returned = 0
            try:
                while True:
                    key, _, _, _ = package1.next()
                    package1.complete(key, made_changes=True, result="Summary")
                    nb_steps_returned += 1
            except graph.NoMorePendingStep:
                pass
            self.assertEqual(nb_steps_returned, nb_pending)

    def test_nonexistent_account(self):
        """Test the behavior with current deployments in AWS accounts that
        are not active anymore.
        """
        package1 = Package()
        package1.full_init(self.orga)
        # Add a fake deployment to the current state
        fake_key = store.ModuleAccountRegionKey(
            "terraform1", "098765432109", "us-east-1"
        )
        package1.current[fake_key] = store.CurrentDeploymentDetails(
            variables={},
            var_from_outputs={},
            dependencies=[],
            module_hash="fake_hash",
            outputs={},
            last_changed_time="",
        )
        package1.save()
        # Reload the package and check that there should be one pending but
        # skipped destroy change, because the fake AWS account ID doesn't
        # belong to the list of active accounts in the organization
        package2 = Package()
        package2.full_init(self.orga)
        export = package2.export_changes()
        self.assertEqual(len(export["PendingButSkippedChanges"]["Destroy"]), 1)

    def test_fail(self):
        """Test the behavior with a failed deployment."""
        # Add a filter to include only the account ID "123456789012" (master)
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"include_account_ids": [self.fake_orga.account_a_id]}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Retrieve the next step and mark it as failed. It should return a step
            # for module "python1"
            key1, _, nb_attempts, _ = package1.next()
            self.assertEqual(nb_attempts, 1)
            package1.fail(key1, result="Failed")
            # Try again, it should raise NoProcessableStep because this step must
            # be completed before other steps can be processed and
            # `DelayBeforeRetrying` is step to 1 second.
            with self.assertRaises(graph.NoProcessableStep):
                package1.next()
            # Retry after 1 seconds
            time.sleep(1)
            key2, _, nb_attempts, _ = package1.next()
            self.assertEqual(key1, key2)
            self.assertEqual(nb_attempts, 2)
            package1.complete(key2, made_changes=True, result="Completed")
            # Retrieve the next step, it should corresponds to a deployment
            # for "terraform1" module
            key, _, _, _ = package1.next()
            self.assertEqual(key.module, "terraform1")

    def test_conditional_update_1(self):
        """Test the behavior with conditional updates and output values that
        don't change.
        """
        package1 = Package()
        package1.full_init(self.orga)
        # Mark all steps as completed
        try:
            while True:
                key, _, _, _ = package1.next()
                package1.complete(
                    key,
                    made_changes=True,
                    result="Summary",
                    outputs={"outputPython1": "value"},
                )
        except graph.NoMorePendingStep:
            pass
        # Change the value of the module hash in the current state to simulate
        # the need to update a deployment
        key = store.ModuleAccountRegionKey(
            "python1", self.fake_orga.account_a_id, "us-east-1"
        )
        package1.current[key].module_hash = "fake_hash"
        # Reload the package
        package1.save()
        package2 = Package()
        package2.full_init(self.orga)
        # Check that there is one update, and one conditional update because
        # the module "terraform1" depends on the outputs of "python1"
        export = package2.export_changes()
        self.assertEqual(len(export["PendingChanges"]["Update"]), 1)
        self.assertEqual(
            len(export["PendingChanges"]["ConditionalUpdate"]),
            len(mock.ENABLED_REGIONS),
        )
        # Mark the first step for "python1" as completed and don't change the
        # value of the outputs
        key, _, _, _ = package2.next()
        package2.complete(
            key,
            made_changes=True,
            result="Summary",
            outputs={"outputPython1": "value"},
        )
        # There should be no other step because the value of outputs have not
        # changed
        with self.assertRaises(graph.NoMorePendingStep):
            package2.next()
        # Check that there are multiple completed conditional updates in the
        # results and none has result in changes
        export = package2.export_results()
        self.assertGreater(len(export["Completed"]["ConditionalUpdate"]), 1)
        self.assertNotIn(
            True,
            [
                item["ResultedInChanges"]
                for item in export["Completed"]["ConditionalUpdate"]
            ],
        )

    def test_conditional_update_2(self):
        """Test the behavior with conditional updates and output values that
        change.
        """
        package1 = Package()
        package1.full_init(self.orga)
        # Mark all steps as completed
        try:
            while True:
                key, _, _, _ = package1.next()
                package1.complete(
                    key,
                    made_changes=True,
                    result="Summary",
                    outputs={"outputPython1": "value"},
                )
        except graph.NoMorePendingStep:
            pass
        # Change the value of the module hash in the current state to simulate
        # the need to update a deployment
        key = store.ModuleAccountRegionKey(
            "python1", self.fake_orga.account_a_id, "us-east-1"
        )
        package1.current[key].module_hash = "fake_hash"
        # Reload the package
        package1.save()
        package2 = Package()
        package2.full_init(self.orga)
        # Check that there is one update, and one conditional update before
        # the module "terraform1" depends on the outputs of "python1"
        export = package2.export_changes()
        self.assertEqual(len(export["PendingChanges"]["Update"]), 1)
        self.assertEqual(
            len(export["PendingChanges"]["ConditionalUpdate"]),
            len(mock.ENABLED_REGIONS),
        )
        # Mark the first step for "python1" as completed and change the value
        # of the output
        key, _, _, _ = package2.next()
        package2.complete(
            key,
            made_changes=True,
            result="Summary",
            outputs={"outputPython1": "new_value"},
        )
        # Check that the value of the current state is correct
        self.assertEqual(package2.current[key].outputs["outputPython1"], "new_value")
        # The next step should be "update" because the outputs value changed,
        # and that the value of the variable matches the value of the
        # ascendant output
        key, action, _, _ = package2.next()
        self.assertEqual(action, "update")
        self.assertEqual(package2.target[key].variables["varTerraform1"], "new_value")
        package2.complete(key, made_changes=True, result="Summary")
        # Check that there is one completed conditional update in the results
        # and it resulted in changes
        export = package2.export_results()
        self.assertEqual(len(export["Completed"]["ConditionalUpdate"]), 1)
        self.assertTrue(
            export["Completed"]["ConditionalUpdate"][0]["ResultedInChanges"]
        )

    def test_no_update_needed(self):
        """Test the behavior when there are no updates to make."""
        package1 = Package()
        package1.full_init(self.orga)
        # Mark all steps as completed
        try:
            while True:
                key, _, _, _ = package1.next()
                package1.complete(key, made_changes=True, result="Summary")
        except graph.NoMorePendingStep:
            pass
        # Reload the package and check that there are no pending changes
        package1.save()
        package2 = Package()
        package2.full_init(self.orga)
        self.assertFalse(package2.analyze_changes())

    # Set the CLI argument `--force-update`
    def test_force_update(self):
        """Test the behavior when there are no changes to make but we force
        updates using CLI argument.
        """
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"force_update": True}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Mark all steps as completed
            try:
                while True:
                    key, _, _, _ = package1.next()
                    package1.complete(key, made_changes=True, result="Summary")
            except graph.NoMorePendingStep:
                pass
            # Reload the package and check that there are pending changes
            package1.save()
            package2 = Package()
            package2.full_init(self.orga)
            self.assertTrue(package2.analyze_changes())

    # Set the CLI command to "preview"
    def test_preview_with_dependencies(self):
        """Test the behavior in "preview" mode where deployments have changes
        to be made, but they dependent on other deployments also with
        pending changes.
        """
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"command": "preview"}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Mark all steps as completed
            try:
                while True:
                    key, _, _, _ = package1.next()
                    package1.complete(key, made_changes=True, result="Summary")
            except graph.NoMorePendingStep:
                pass
            # Check that there is at least one deployment that failed, because
            # changes to resources cannot be previewed when dependencies have
            # pending changes
            export = package1.export_results()
            self.assertGreaterEqual(len(export["Failed"]["Create"]), 1)

    def test_missing_dependencies_not_ignored_1(self):
        """Test the behavior with dependencies that don't exist with the attribute
        `IgnoreIfNotExists` not set.
        """
        # Use the package file "package3.yaml" for this test
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"package_file": mock.get_test_path("package3.yaml")}),
        ):
            package = Package()
            with self.assertRaises(graph.GraphError):
                package.full_init(self.orga)

    def test_missing_dependencies_not_ignored_2(self):
        """Test the behavior with dependencies that don't exist with the attribute
        `IgnoreIfNotExists` not set for all dependencies.
        """
        # Use the package file "package3.yaml" for this test
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"package_file": mock.get_test_path("package3.yaml")}),
        ):
            package = Package()
            # Modify the package content to set "IgnoreIfNotExists" for the
            # first deployment block only
            deployments = package.package["Modules"]["terraform1"]["Deployments"]
            deployments[0]["Dependencies"][0]["IgnoreIfNotExists"] = True
            with self.assertRaises(graph.GraphError):
                package.full_init(self.orga)

    def test_missing_dependencies_ignored(self):
        """Test the behavior with dependencies that don't exist with the attribute
        `IgnoreIfNotExists` set to True for all dependencies.
        """
        # Use the package file "package3.yaml" for this test
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"package_file": mock.get_test_path("package3.yaml")}),
        ):
            package = Package()
            # Modify the package content to set "IgnoreIfNotExists" for both
            # deployment blocks
            deployments = package.package["Modules"]["terraform1"]["Deployments"]
            deployments[0]["Dependencies"][0]["IgnoreIfNotExists"] = True
            deployments[1]["VariablesFromOutputs"]["var1"]["IgnoreIfNotExists"] = True
            package.full_init(self.orga)
            # Mark all steps as completed
            try:
                while True:
                    key, _, _, _ = package.next()
                    package.complete(key, made_changes=True, result="Summary")
            except graph.NoMorePendingStep:
                pass
            # Check that the value of the "var1" for the module "terraform1"
            # is still equal to "value1" as the "VariablesFromOutputs"
            # dependency does not exist
            key = store.ModuleAccountRegionKey(
                "terraform1", "123456789012", "us-east-2"
            )
            self.assertEqual(package.current[key].variables["var1"], "value1")

    def test_package_save_state_enabled(self):
        """Test the behavior with the argument `save-state-every-seconds` = 1."""
        with patch(
            "aws_orga_deployer.config.CLI",
            update_cli_filters({"save_state_every_seconds": 1}),
        ):
            package1 = Package()
            package1.full_init(self.orga)
            # Mark all steps as completed
            try:
                while True:
                    key, _, _, _ = package1.next()
                    package1.complete(key, made_changes=True, result="Summary")
            except graph.NoMorePendingStep:
                pass
            time.sleep(2)
            package2 = Package()
            package2.full_init(self.orga)
            export = package2.export_changes()
            self.assertNotIn("PendingChanges", export.keys())
            package1.save(stop_autosave=True)
            package2.save(stop_autosave=True)

    def test_package_save_state_disabled(self):
        """Test the behavior without the argument `save-state-every-seconds`."""
        package1 = Package()
        package1.full_init(self.orga)
        # Mark all steps as completed
        try:
            while True:
                key, _, _, _ = package1.next()
                package1.complete(key, made_changes=True, result="Summary")
        except graph.NoMorePendingStep:
            pass
        # First attempt without saving
        time.sleep(2)
        package2 = Package()
        package2.full_init(self.orga)
        export = package2.export_changes()
        self.assertIn("PendingChanges", export.keys())
        # Second attempt after saving
        package1.save()
        package3 = Package()
        package3.full_init(self.orga)
        export = package3.export_changes()
        self.assertNotIn("PendingChanges", export.keys())

    def test_remove_orphans(self):
        """Test that remove_orphans removes module deployments for accounts
        that no longer exist in the organization.
        """
        package1 = Package()
        package1.full_init(self.orga)
        # Add a fake deployment to the current state
        fake_key = store.ModuleAccountRegionKey(
            "terraform1", "098765432109", "us-east-1"
        )
        package1.current[fake_key] = store.CurrentDeploymentDetails(
            variables={},
            var_from_outputs={},
            dependencies=[],
            module_hash="fake_hash",
            outputs={},
            last_changed_time="",
        )
        package1.save()
        # Reload the package and remove orphans
        package2 = Package()
        package2.full_init(self.orga)
        orphans_removed = package2.remove_orphans()
        self.assertIn(fake_key.to_dict(), orphans_removed)
        # Reload the package and check that there are no module deployments
        # to destroy
        package3 = Package()
        package3.full_init(self.orga)
        export = package3.export_changes()
        self.assertNotIn("PendingButSkippedChanges", export.keys())


class TestInvalidPackage(unittest.TestCase):
    """Test the class Package with invalid package definition files."""

    @mock_s3
    def test_invalid_package_1(self):
        """Test the behavior with an empty package definition file."""
        # Create a fake bucket
        mock.create_fake_bucket()
        # Change the location of the package definition file
        mock.mock_cli_arguments(package_filename="invalid-package1.yaml")
        with self.assertRaises(PackageError):
            Package()

    @mock_s3
    def test_invalid_package_2(self):
        """Test the behavior with an invalid package definition file."""
        # Create a fake bucket
        mock.create_fake_bucket()
        # Change the location of the package definition file
        mock.mock_cli_arguments(package_filename="invalid-package2.yaml")
        with self.assertRaises(PackageError):
            Package()



"""Test the module `aws_orga_deployer.utils`."""

# COMPLETED
import unittest

from aws_orga_deployer import utils


class TestExecMultithread(unittest.TestCase):
    """Test the function `exec_multithread`."""

    def test_exec_multithread_ok(self):
        """Check that the function `exec_multithread` works as expected."""
        input_list = range(10)
        result = []

        def func(value):
            result.append(value)

        utils.exec_multithread(input_list, func, 5)
        self.assertCountEqual(input_list, result)

    def test_exec_multithread_exception(self):
        """Check that exception raised by the threads are propagated to the
        parent thread.
        """

        def func(value):
            raise RuntimeError()

        with self.assertRaises(RuntimeError):
            utils.exec_multithread(range(10), func, 5)



"""Modules for each supported engine."""

import os
from importlib import import_module
from typing import Any, Dict, List

# List the supported engines
engine_names: List[str] = [
    os.path.splitext(file)[0]
    for file in os.listdir(os.path.dirname(os.path.abspath(__file__)))
    if not file in ("__init__.py", "base.py") and file.endswith(".py")
]


# Load the class Engine of each supported module in a dict
ENGINES: Dict[str, Any] = {}
for engine_name in engine_names:
    module = import_module(f"aws_orga_deployer.engines.{engine_name}")
    ENGINES[engine_name] = module.Engine



"""Load the package definition file and determine the modules to deploy."""

import json
import logging
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple

import yaml

from aws_orga_deployer import config
from aws_orga_deployer.orga import OrgaParser
from aws_orga_deployer.package import schema
from aws_orga_deployer.package.graph import DeploymentGraph
from aws_orga_deployer.package.store import (
    CurrentDeploymentDetails,
    CurrentStateStore,
    ModuleAccountRegionKey,
    TargetDeploymentDetails,
)

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class PackageError(Exception):
    """Exception raised when the package definition file is invalid."""

    def __init__(self, message: str) -> None:
        self.message = f"The package definition YAML file is invalid - {message}"
        super().__init__(self.message)


def _replace_keywords(source: Any, current_account_id: str, current_region: str) -> Any:
    """Replace any occurences of `${CURRENT_ACCOUNT_ID}` by `current_account_id`
    and of `${CURRENT_REGION}` by `current_region` recursively across lists and
    dictionaries.

    Args:
        source: Input object
        current_account_id (str): Value to replace `${CURRENT_ACCOUNT_ID}`
        current_region (str): Value to replace `${CURRENT_REGION}`
    """
    if isinstance(source, str):
        source = source.replace("${CURRENT_ACCOUNT_ID}", current_account_id)
        source = source.replace("${CURRENT_REGION}", current_region)
        return source
    if isinstance(source, dict):
        return {
            key: _replace_keywords(value, current_account_id, current_region)
            for key, value in source.items()
        }
    if isinstance(source, list):
        return [
            _replace_keywords(item, current_account_id, current_region)
            for item in source
        ]
    return source


class Package:
    """Class to load and parse the package definition file, determine the
    modules deployed and the changes to be made.
    """

    orga: OrgaParser
    current: CurrentStateStore
    target: Dict[ModuleAccountRegionKey, TargetDeploymentDetails]
    graph: DeploymentGraph
    cli_filters: Dict[str, Set]
    modules_config: Dict[str, Any]

    def __init__(self):
        """Load and validate the package definition file. This is needed to
        retrieve package configuration settings, such as the name of the bucket
        that stores persistent data.
        """
        # Attributes that will be populated by `full_init`
        self.orga = None
        self.current = None
        self.target = None
        self.graph = None
        self.cli_filters = None
        self.modules_config = None
        # Load the YAML file and validate its schema
        # pylint: disable=broad-exception-caught
        with open(config.CLI["package_file"], "r", encoding="utf-8") as stream:
            try:
                self.package = yaml.safe_load(stream)
                schema.validate(self.package)
            except Exception as err:
                raise PackageError(err) from err
        # Store the package configuration for use by other modules (attribute
        # `PackageConfiguration` of the YAML file)
        config.PACKAGE = self.package["PackageConfiguration"]

    def full_init(self, orga: OrgaParser) -> None:
        """Load additional elements needed for the CLI command `list`, `preview`
        or `deploy`: Determine the module deployments that should exist in the
        target state, the module deployments that currently exist by reading the
        package state file, and evaluate the changes to be made.

        Args:
            orga: Object to query for information on the accounts and the
                organizational units.
        """
        self.orga = orga
        self.target = {}
        self.modules_config = {}
        auto_save_period = config.CLI.get("save_state_every_seconds", 0)
        self.current = CurrentStateStore(period=auto_save_period)
        self.graph = DeploymentGraph()
        self._init_target_deployments()
        self._init_cli_filters()
        self._init_graph()

    def _init_target_deployments(self) -> None:
        """Determine the module deployments that must exist in the target
        state based on the package definition file.
        """
        for module, module_block in self.package["Modules"].items():
            # Determine the module configuration parameters for all deployments
            # of this module
            engine = config.MODULES[module].engine
            module_config = {}
            if "DefaultModuleConfiguration" in self.package:
                default_config = self.package["DefaultModuleConfiguration"]
                module_config.update(default_config.get("All", {}))
                module_config.update(default_config.get(engine, {}))
            module_config.update(module_block.get("Configuration", {}))
            # Validate the module configuration
            try:
                config.MODULES[module].validate_module_config(module_config)
            except AssertionError as err:
                raise PackageError(f"Configuration of {module}: {err}") from err
            self.modules_config[module] = module_config
            # For each deployment block, i.e. each item in the attribute
            # `Deployments`
            for deployment_block in module_block["Deployments"]:
                self._process_deployment_block(deployment_block, module)

    def _process_deployment_block(
        self, deployment_block: Dict[str, Any], module: str
    ) -> None:
        """Process a deployment block, i.e. an item in the attribute
        `Deployments` of a module in the package definition file, to determine
        the expected module deployments.

        Args:
            deployment_block: Content of the deployment block
            module: Module name
        """
        engine = config.MODULES[module].engine
        module_block = self.package["Modules"][module]
        # Determine the variables
        variables_block = {}
        if "DefaultVariables" in self.package:
            default_variables = self.package["DefaultVariables"]
            variables_block.update(default_variables.get("All", {}))
            variables_block.update(default_variables.get(engine, {}))
        variables_block.update(module_block.get("Variables", {}))
        variables_block.update(deployment_block.get("Variables", {}))
        # Determine the dependencies
        dependencies_block = deployment_block.get("Dependencies", [])
        var_from_outputs_block = {}
        var_from_outputs_block.update(module_block.get("VariablesFromOutputs", {}))
        var_from_outputs_block.update(deployment_block.get("VariablesFromOutputs", {}))
        # Identify the list of accounts where to deploy the module for this
        # block. Filter or exclude accounts based on the `Include` or `Exclude`
        # attributes
        account_ids = set(self.orga.get_all_accounts())
        include = deployment_block.get("Include", {})
        if "AccountIds" in include:
            tmp = self.orga.get_accounts_by_id(include["AccountIds"])
            account_ids = account_ids & set(tmp)
        if "AccountNames" in include:
            tmp = self.orga.get_accounts_by_name(include["AccountNames"])
            account_ids = account_ids & set(tmp)
        if "AccountTags" in include:
            tmp = self.orga.get_accounts_by_tag(include["AccountTags"])
            account_ids = account_ids & set(tmp)
        if "OUIds" in include:
            tmp = self.orga.get_accounts_by_ou(include["OUIds"])
            account_ids = account_ids & set(tmp)
        if "OUTags" in include:
            tmp = self.orga.get_accounts_by_ou_tag(include["OUTags"])
            account_ids = account_ids & set(tmp)
        exclude = deployment_block.get("Exclude", {})
        if "AccountIds" in exclude:
            tmp = self.orga.get_accounts_by_id(exclude["AccountIds"])
            account_ids = account_ids - set(tmp)
        if "AccountNames" in exclude:
            tmp = self.orga.get_accounts_by_name(exclude["AccountNames"])
            account_ids = account_ids - set(tmp)
        if "AccountTags" in exclude:
            tmp = self.orga.get_accounts_by_tag(exclude["AccountTags"])
            account_ids = account_ids - set(tmp)
        if "OUIds" in exclude:
            tmp = self.orga.get_accounts_by_ou(exclude["OUIds"])
            account_ids = account_ids - set(tmp)
        if "OUTags" in exclude:
            tmp = self.orga.get_accounts_by_ou_tag(exclude["OUTags"])
            account_ids = account_ids - set(tmp)
        # For each account in the scope of the deployment block
        for account_id in account_ids:
            # Identify the regions where the module should be deployed. Start
            # from all enabled regions in the account and include and exclude
            # regions as defined by CLI arguments
            regions = set(self.orga.get_account_regions(account_id, ["ALL_ENABLED"]))
            if deployment_block.get("Include", {}).get("Regions"):
                tmp = self.orga.get_account_regions(
                    account_id, deployment_block["Include"]["Regions"]
                )
                regions = regions & set(tmp)
            if deployment_block.get("Exclude", {}).get("Regions"):
                tmp = self.orga.get_account_regions(
                    account_id, deployment_block["Exclude"]["Regions"]
                )
                regions = regions - set(tmp)
            # For each region targeted in the deployment scope for this account
            for region in regions:
                # Replace the keywords `${CURRENT_REGION}` and
                # `${CURRENT_ACCOUNT_ID}` by the current region and account ID.
                # The variables ending with `_block` are deep copied because
                # the value `${CURRENT_REGION}` is replaced by the current
                # region, so we cannot use the same objects for all regions.
                dependencies = _replace_keywords(
                    deepcopy(dependencies_block), account_id, region
                )
                var_from_outputs = _replace_keywords(
                    deepcopy(var_from_outputs_block), account_id, region
                )
                variables = _replace_keywords(
                    deepcopy(variables_block), account_id, region
                )
                # Add the target deployment
                key = ModuleAccountRegionKey(module, account_id, region)
                self.target[key] = TargetDeploymentDetails(
                    variables=variables,
                    var_from_outputs=var_from_outputs,
                    dependencies=dependencies,
                    module_hash=config.MODULES[module].module_hash,
                )
                # Set the value the variables that depend on the outputs of
                # other module deployments
                self._set_variables_from_outputs(key)

    def _set_variables_from_outputs(self, t_key: ModuleAccountRegionKey) -> None:
        """For a given module deployment, set the value of its variables that
        depend on the outputs of other module deployments.

        Args:
            t_key: Step key for which variables must be updated with values
                of ascendant outputs
        """
        # Continue only if the module deployment exists in the current state
        # (not for "destroy")
        if not t_key in self.target:
            return
        t_content = self.target[t_key]
        # For each variable valued from the outputs of other module deployments
        for var_name, var_from in t_content.var_from_outputs.items():
            from_key = ModuleAccountRegionKey(
                var_from["Module"],
                var_from["AccountId"],
                var_from["Region"],
            )
            c_content = self.current.get(from_key)
            # If the source module deployment exists and has outputs, retrieve
            # the value of the output and modify the value of the variables
            if not c_content is None:
                if var_from["OutputName"] in c_content.outputs:
                    output_value = c_content.outputs.get(var_from["OutputName"])
                    t_content.variables[var_name] = output_value

    def _init_cli_filters(self) -> None:
        """Identify the filters defined by the CLI arguments `--include`
        and `--exclude` that restrict the modules, accounts or regions to
        deploy.
        """
        # Identify the deployments that the CLI arguments allow to create,
        # update or delete. Default to all existing modules.
        modules = set([*config.MODULES])
        if not config.CLI.get("include_modules") is None:
            modules = modules & set(config.CLI["include_modules"])
        if not config.CLI.get("exclude_modules") is None:
            modules = modules - set(config.CLI["exclude_modules"])
        # Identify the accounts that the CLI arguments allow to create,
        # update or delete. Default to all AWS accounts.
        account_ids = set(self.orga.get_all_accounts())
        if not config.CLI.get("include_account_ids") is None:
            tmp = self.orga.get_accounts_by_id(config.CLI["include_account_ids"])
            account_ids = account_ids & set(tmp)
        if not config.CLI.get("include_account_names") is None:
            tmp = self.orga.get_accounts_by_name(config.CLI["include_account_names"])
            account_ids = account_ids & set(tmp)
        if not config.CLI.get("include_account_tags") is None:
            tmp = self.orga.get_accounts_by_tag(config.CLI["include_account_tags"])
            account_ids = account_ids & set(tmp)
        if not config.CLI.get("include_ou_ids") is None:
            tmp = self.orga.get_accounts_by_ou(config.CLI["include_ou_ids"])
            account_ids = account_ids & set(tmp)
        if not config.CLI.get("include_ou_tags") is None:
            tmp = self.orga.get_accounts_by_ou_tag(config.CLI["include_ou_tags"])
            account_ids = account_ids & set(tmp)
        if not config.CLI.get("exclude_account_ids") is None:
            tmp = self.orga.get_accounts_by_id(config.CLI["exclude_account_ids"])
            account_ids = account_ids - set(tmp)
        if not config.CLI.get("exclude_account_names") is None:
            tmp = self.orga.get_accounts_by_name(config.CLI["exclude_account_names"])
            account_ids = account_ids - set(tmp)
        if not config.CLI.get("exclude_account_tags") is None:
            tmp = self.orga.get_accounts_by_tag(config.CLI["exclude_account_tags"])
            account_ids = account_ids - set(tmp)
        if not config.CLI.get("exclude_ou_ids") is None:
            tmp = self.orga.get_accounts_by_ou(config.CLI["exclude_ou_ids"])
            account_ids = account_ids - set(tmp)
        if not config.CLI.get("exclude_ou_tags") is None:
            tmp = self.orga.get_accounts_by_ou_tag(config.CLI["exclude_ou_tags"])
            account_ids = account_ids - set(tmp)
        # Identify the regions that the CLI arguments allow to create,
        # update or delete. Default to all enabled regions.
        regions = set(self.orga.get_all_enabled_regions())
        if not config.CLI.get("include_regions") is None:
            regions = regions & set(config.CLI["include_regions"])
        if not config.CLI.get("exclude_regions") is None:
            regions = regions - set(config.CLI["exclude_regions"])
        # Store resulting filters in a dict
        self.cli_filters = {
            "Modules": modules,
            "AccountIds": account_ids,
            "Regions": regions,
        }

    def _init_graph(self) -> None:
        self._add_graph_steps()
        self._add_graph_dependencies()
        self.graph.validate()

    def _add_graph_steps(self) -> None:
        # For each module deployment expected in the target state, check if
        # it must be created or updated
        for t_key, t_content in self.target.items():
            c_content = self.current.get(t_key)
            skip = self._is_skipped_by_cli_filters(t_key)
            # If the module deployment already exists, check if it must be
            # updated
            if not c_content is None:
                if self._check_update_needed(c_content, t_content):
                    action = "update"
                else:
                    action = "none"
            else:
                action = "create"
            max_attempts, delay = self._get_retry_parameters(t_key.module)
            self.graph.add_step(t_key, action, skip, max_attempts, delay)
        # For each module deployment that currently exists, check if it must be
        # deleted, i.e. if it should not exist in the target state
        for c_key, c_content in self.current.items():
            # When there are deployments to destroy, there must be a block in
            # the package definition file for this module to retrieve module
            # configuration parameters
            if not c_key.module in self.modules_config:
                raise PackageError(
                    f'There must be a block for the module "{c_key.module}" even with'
                    " an empty list of deployments"
                )
            if not c_key in self.target:
                skip = self._is_skipped_by_cli_filters(c_key)
                max_attempts, delay = self._get_retry_parameters(c_key.module)
                self.graph.add_step(c_key, "destroy", skip, max_attempts, delay)

    def _get_retry_parameters(self, module: str) -> Tuple[int, int]:
        """Return the retry parameters for a given module from the module
        configuration parameters.

        Args:
            module: Module name

        Returns:
            tuple:
                Maximum number of attempts. Default is 1
                Number of seconds to wait before retrying. Default is 0
        """
        if (
            not module in self.modules_config
            or not "Retry" in self.modules_config[module]
        ):
            return 1, 0
        max_attempts = self.modules_config[module]["Retry"].get("MaxAttempts", 1)
        delay = self.modules_config[module]["Retry"].get("DelayBeforeRetrying", 0)
        return max_attempts, delay

    def _check_update_needed(
        self, current: CurrentDeploymentDetails, target: TargetDeploymentDetails
    ) -> bool:
        """Return True if the module deployment must be updated because the
        current and target state differs, and because the CLI argument
        `--force-update` is set.

        Args:
            current: Current deployment details
            target: Target deployment details

        Returns:
            True if update is needed
        """
        return config.CLI.get("force_update", False) is True or not (
            current.module_hash == target.module_hash
            and current.variables == target.variables
        )

    def _add_graph_dependencies(self) -> None:
        # For each step in the graph, add dependencies as graph edges
        for to_key in self.graph.list_steps():
            # Get the corresponding module deployment details, either from the
            # list of the target deployments or current deployments
            to_details = self.target.get(to_key)
            if to_details is None:
                to_details = self.current[to_key]
            # Add dependencies defined by the `Dependencies` and/or
            # `VariablesFromOutputs` attributes. If a dependency is both in
            # `Dependencies` and `VariablesFromOutputs`, the latter prevails
            for dependency in to_details.dependencies:
                from_key = ModuleAccountRegionKey(
                    dependency["Module"],
                    dependency["AccountId"],
                    dependency["Region"],
                )
                ignore_if_not_exists = bool(dependency.get("IgnoreIfNotExists", False))
                self.graph.add_dependency(
                    from_key,
                    to_key,
                    is_var=False,
                    ignore_if_not_exists=ignore_if_not_exists,
                )
            for dependency in to_details.var_from_outputs.values():
                from_key = ModuleAccountRegionKey(
                    dependency["Module"],
                    dependency["AccountId"],
                    dependency["Region"],
                )
                ignore_if_not_exists = bool(dependency.get("IgnoreIfNotExists", False))
                self.graph.add_dependency(
                    from_key,
                    to_key,
                    is_var=True,
                    ignore_if_not_exists=ignore_if_not_exists,
                )

    def _is_skipped_by_cli_filters(self, key: ModuleAccountRegionKey) -> bool:
        """Check if a given module deployment should be skipped due to the CLI
        arguments.

        Args:
            key: Step key.

        Returns:
            True if skipped
        """
        return not (
            key.module in self.cli_filters["Modules"]
            and key.account_id in self.cli_filters["AccountIds"]
            and key.region in self.cli_filters["Regions"]
        )

    def next(self) -> Tuple[ModuleAccountRegionKey, str, int, int]:
        """Retrieve the next deployment to process and the action to make.
        Also update the value of the variables that eventually depends on the
        outputs of other module deployments.

        Returns:
            tuple:
                Step key
                Action to make. "conditional-update" is replaced by
                    "skipped" or "update" depending on whether output values
                    changed or not
                Current attempt number
                Maximum number of attempts

        Raises:
            aws_orga_deployer.package.graph.NoProcessableStep
            aws_orga_deployer.package.graph.NoMorePendingStep
        """
        while True:
            key = self.graph.next()
            details = self.graph.get_step_details(key)
            # If the command is "preview" and the step is dependent on other
            # steps with pending changes, changes to resources cannot be
            # previewed as they depend on a state that doesn't exist yet (e.g.
            # value of the dependent outputs). Therefore, the step is failed.
            if (
                config.CLI["command"] == "preview"
                and self.graph.has_ascendants_with_changes(key)
                and details.action != "destroy"
            ):
                message = (
                    "Unable to preview changes as this deployment is dependent on other"
                    " deployments with pending changes"
                )
                self.graph.fail(key, message)
                LOGGER.error("%s %s", key, message)
                continue
            # Update the variables with the latest values of the ascendant
            # outputs
            self._set_variables_from_outputs(key)
            if not details.action == "conditional-update":
                return key, details.action, details.nb_attempts, details.max_attempts
            # If the action is "conditional-update" and the value of the values
            # of the outputs on which depend the deployment have not changed,
            # skip the deployment
            current = self.current[key]
            target = self.target[key]
            if self._check_update_needed(current, target):
                return key, "update", details.nb_attempts, details.max_attempts
            self.graph.complete(
                key,
                False,
                "No changes required because the dependent output values have not"
                " changed",
            )

    def get_module_config(self, key: ModuleAccountRegionKey) -> Dict[str, Any]:
        """Return the module configuration for a given module deployment. The
        keywords `${CURRENT_ACCOUNT_ID}` and `${CURRENT_ACCOUNT}` are replaced
        by the account ID and region of the key.

        Args:
            key: Step key.

        Returns:
            Module configuration parameters
        """
        return _replace_keywords(
            deepcopy(self.modules_config[key.module]), key.account_id, key.region
        )

    def complete(
        self,
        key: ModuleAccountRegionKey,
        made_changes: bool,
        result: str,
        detailed_results: Optional[Dict[str, Any]] = None,
        outputs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Mark a deployment as completed, and update the current package state
        if the command is "apply".

        Args:
            key: Step key
            made_changes: True if the step resulted in changes made or to be made
            result: One-line summary of the result.
            detailed_results: Dict with detailed results.
            outputs: Outputs returned by the module deployment, only when the
                action is not "destroy". Default is None to store an empty dict.
        """
        self.graph.complete(key, made_changes, result, detailed_results)
        # Update the current state store only if the command is "apply"
        if config.CLI["command"] == "apply":
            details = self.graph.get_step_details(key)
            # If the action is "destroy", remove it from the current state,
            # otherwise update the current state from the target state
            if details.action == "destroy":
                del self.current[key]
            else:
                self.current[key] = CurrentDeploymentDetails(
                    variables=self.target[key].variables,
                    var_from_outputs=self.target[key].var_from_outputs,
                    dependencies=self.target[key].dependencies,
                    module_hash=self.target[key].module_hash,
                    outputs=outputs if not outputs is None else {},
                    last_changed_time=str(datetime.utcnow()),
                )

    def fail(
        self,
        key: ModuleAccountRegionKey,
        result: str,
        detailed_results: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Mark a deployment as failed.

        Args:
            key: Step key.
            result: One-line summary of the result.
            detailed_results: Dict with detailed results.
        """
        self.graph.fail(key, result, detailed_results)

    def update_hash(self, key: ModuleAccountRegionKey) -> bool:
        """Update the value of the module hash if the deployment action
        is "update". Then, mark the deployment as completed.

        Args:
            key: Step key.

        Returns:
            True if the module hash was updated, False otherwise
        """
        details = self.graph.get_step_details(key)
        if details.action == "update":
            if self.current[key].module_hash != self.target[key].module_hash:
                self.current[key].module_hash = self.target[key].module_hash
                self.current[key].last_changed_time = str(datetime.utcnow())
                self.graph.complete(key, True, "Updated the module hash")
                return True
        self.graph.complete(key, False, "No action needed")
        return False

    def save(self, stop_autosave: bool = False) -> None:
        """Save the current package state. Optionally stop the auto-save of the
        current state  That is particularly useful for testing because
        the auto-save might continue to write to S3 after the test has
        completed.

        Args:
            stop_autosave (bool): Stop the daemon thread of the current state
                store that saves the state to S3 at regular intervals.
        """
        self.current.save()
        if stop_autosave:
            self.current.stop()

    def analyze_changes(self) -> bool:
        """Display information about the changes to make and return whether
        there are pending changes to be made.

        Returns:
            bool: True if there are pending changes to be made.
        """
        result = {
            "create": {"pending": 0, "pending_skipped": 0},
            "update": {"pending": 0, "pending_skipped": 0},
            "conditional-update": {"pending": 0, "pending_skipped": 0},
            "destroy": {"pending": 0, "pending_skipped": 0},
        }
        # Count the number of changes to make
        pending_changes = 0
        for _, step in self.graph.list_steps_with_details():
            if step.action == "none":
                continue
            if step.skip:
                result[step.action]["pending_skipped"] += 1
            else:
                pending_changes += 1
                result[step.action]["pending"] += 1
        # Print the summary
        for action, prefix in (
            ("create", "Deployments to create"),
            ("update", "Deployments to update"),
            (
                "conditional-update",
                (
                    "Deployments thay may need updates if the outputs on which they"
                    " depend change"
                ),
            ),
            ("destroy", "Deployments to destroy"),
        ):
            if result[action]["pending"] + result[action]["pending_skipped"] > 0:
                message = prefix + ": %s (%s skipped due to CLI filters)"
                LOGGER.info(
                    message,
                    result[action]["pending"],
                    result[action]["pending_skipped"],
                )
        if pending_changes == 0:
            LOGGER.info("No changes to be made during this run")
        # Return True if there are pending changes that are not skipped
        return pending_changes > 0

    def export_changes(self) -> Dict[str, Any]:
        """Export the list of module deployments and changes to be made.

        Returns:
            dict: Dictionary whose structure is::

                {
                    "PendingChanges|PendingButSkippedChanges": {
                        "Create": [
                            {
                                "Deployment": {
                                    "Module" (str),
                                    "AccountId" (str),
                                    "AccountName" (str),
                                    "Region" (str)
                                },
                                "ModuleConfig" (dict): Module configuration,
                                "TargetState" (dict): Details about target state,
                            }
                        ],
                        "Update|ConditionalUpdate": [
                            {
                                "Deployment": {
                                    "Module" (str),
                                    "AccountId" (str),
                                    "AccountName" (str),
                                    "Region" (str)
                                },
                                "ModuleConfig" (dict): Module configuration,
                                "CurrentState" (dict): Details about current state,
                                "TargetState" (dict): Details about target state,
                            }
                        ],
                        "Destroy": [
                            {
                                "Deployment": {
                                    "Module" (str),
                                    "AccountId" (str),
                                    "AccountName" (str),
                                    "Region" (str)
                                },
                                "ModuleConfig" (dict): Module configuration,
                                "CurrentState" (dict): Details about current state,
                            }
                        ]
                    },
                    "NoChanges": [
                        {
                            "Deployment": {
                                "Module" (str),
                                "AccountId" (str),
                                "AccountName" (str),
                                "Region" (str)
                            },
                            "ModuleConfig" (dict): Module configuration,
                            "CurrentState" (dict): Details about current state,
                        }
                    ]
                }

        """

        def add_change(
            list_to_append: List,
            step_key: ModuleAccountRegionKey,
            show_current: bool,
            show_target: bool,
        ) -> None:
            """Add a description of a module deployment to the list
            `list_to_append` with details about current and/or target details.

            Args:
                list_to_append: Pointer to the list to which an item must be
                    added.
                key: Step key.
                show_current: True to include details of current state
                show_target: True to include details of target state
            """
            item: Dict[str, Any] = {}
            item["Deployment"] = key.to_dict()
            account_name = self.orga.get_account_name(key.account_id)
            item["Deployment"]["AccountName"] = account_name
            item["ModuleConfig"] = self.get_module_config(step_key)
            if show_current:
                item["CurrentState"] = self.current[step_key].to_dict()
            if show_target:
                item["TargetState"] = self.target[step_key].to_dict()
            list_to_append.append(item)

        result: Dict[str, Any] = {}
        for key, details in self.graph.list_steps_with_details():
            if details.action == "none":
                result.setdefault("NoChanges", [])
                add_change(result["NoChanges"], key, True, False)
                continue
            categ = "PendingButSkippedChanges" if details.skip else "PendingChanges"
            result.setdefault(categ, {})
            if details.action == "create":
                result[categ].setdefault("Create", [])
                add_change(result[categ]["Create"], key, False, True)
            elif details.action == "update":
                result[categ].setdefault("Update", [])
                add_change(result[categ]["Update"], key, True, True)
            elif details.action == "conditional-update":
                result[categ].setdefault("ConditionalUpdate", [])
                add_change(result[categ]["ConditionalUpdate"], key, True, True)
            elif details.action == "destroy":
                result[categ].setdefault("Destroy", [])
                add_change(result[categ]["Destroy"], key, True, False)
        return result

    def analyze_results(self) -> Tuple[bool, bool]:
        """Display information about the results and return changes to
        resources must be made or were made.

        Returns:
            Tuple:
                bool: True if changes to resources have to be made or were made.
                bool: True if some steps failed
        """
        completed = 0
        completed_with_changes = 0
        failed = 0
        pending = 0
        for _, step in self.graph.list_steps_with_details():
            if step.status == "completed":
                completed += 1
                if step.made_changes is True:
                    completed_with_changes += 1
            if step.status == "failed":
                failed += 1
            if step.status == "completed":
                pending += 0
        # Print the summary
        LOGGER.info(
            "%s deployments completed, %s failed, %s still pending",
            completed,
            failed,
            pending,
        )
        return completed_with_changes > 0, failed > 0

    def export_results(self) -> Dict[str, Any]:
        """Export the results of the package execution.

        Returns:
            dict: Dictionary whose structure is::

                {
                    "Completed|Failed|Pending": {
                        "Create|Update|ConditionalUpdate|Destroy": [
                            {
                                "Deployment": {
                                    "Module" (str),
                                    "AccountId" (str),
                                    "AccountName" (str),
                                    "Region" (str)
                                },
                                "NbAttempts" (str),
                                "Result" (str, optional),
                                "DetailedResults" (dict, optional),
                                "ResultedInChanges" (bool): Only for completed steps
                                "Outputs" (dict, optional): Only for completed steps
                                    when command is "apply" and action is not "destroy"
                            }
                        ]
                    }
                }

        """
        result: Dict[str, Any] = {}
        mapping_action = {
            "create": "Create",
            "update": "Update",
            "conditional-update": "ConditionalUpdate",
            "destroy": "Destroy",
        }
        for key, details in self.graph.list_steps_with_details():
            # Do not include deployments that were skipped in the results
            if details.status == "skipped":
                continue
            status = details.status.capitalize()
            action = mapping_action[details.action]
            result.setdefault(status, {})
            result[status].setdefault(action, [])
            item: Dict[str, Any] = {}
            item["Deployment"] = key.to_dict()
            account_name = self.orga.get_account_name(key.account_id)
            item["Deployment"]["AccountName"] = account_name
            item["NbAttempts"] = details.nb_attempts
            if details.result != "":
                item["Result"] = details.result
            if not details.detailed_results is None:
                item["DetailedResults"] = details.detailed_results
            # Only add `ResultedInChanges` if the step has completed
            if details.status == "completed":
                item["ResultedInChanges"] = details.made_changes
                # Only add `Outputs` if the step has completed and the
                # action is not "destroy", i.e. the deployment still exists
                # in the current state store
                if config.CLI["command"] == "apply" and key in self.current:
                    item["Outputs"] = self.current[key].outputs
            result[status][action].append(item)
        return result

    def remove_orphans(self, dry_run: bool = False) -> List[Dict[str, Any]]:
        """Remove the orphaned module deployments from the current state, which
        correspond to the module deployments for the accounts that not longer
        exist in the organization, or regions that are not longer enabled in
        the account.

        Args:
            dry_run: Set to True to return the list of orphans without making
                any changes.

        Returns:
            list: List of dictionaries of the following structure::

                {
                    "Module" (str),
                    "AccountId" (str),
                    "Region" (str)
                }

        """
        orphans_removed: List[Dict[str, Any]] = []
        keys = deepcopy(self.current.keys())
        for key in keys:
            if not self.orga.account_region_exists(key.account_id, key.region):
                orphans_removed.append(key.to_dict())
                if not dry_run:
                    del self.current[key]
        if dry_run:
            LOGGER.info(
                "Found %s orphaned module deployments to remove",
                len(orphans_removed),
            )
        else:
            self.current.save()
            LOGGER.info(
                "Removed %s orphaned module deployments",
                len(orphans_removed),
            )
        return orphans_removed



"""Parse command-line interface arguments."""

import argparse
import logging
import re
import sys
from os import path

from aws_orga_deployer import config


def _account_id_type(value: str) -> str:
    """Check that the value is an account ID."""
    pattern = re.compile(r"^[0-9]{12}$")
    if not pattern.match(value):
        raise argparse.ArgumentTypeError("Invalid format. Must be a 12-digit string")
    return value


def _tag_type(value: str) -> str:
    """Check that the value is in the form TAG_KEY=TAG_VALUE."""
    pattern = re.compile(r"^.+=.+$")
    if not pattern.match(value):
        raise argparse.ArgumentTypeError("Invalid format. Must be TAG_KEY=TAG_VALUE")
    return value


def _check_positive_int(value: str) -> int:
    """Check that the argument value is an integer larger than 0."""
    try:
        int_value = int(value)
        if int_value <= 0:
            raise argparse.ArgumentTypeError(f"{value} must be larger than zero")
    except ValueError as err:
        raise argparse.ArgumentTypeError(f"{value} is not an integer") from err
    return int_value


def _parse_cli_args() -> None:
    """Parse the CLI arguments and store them in the variable `config.CLI`."""

    # Iniialize the parser and define global arguments
    parser = argparse.ArgumentParser(
        prog="aws-orga-deployer",
        description=(
            "AWS Orga Deployer: Deploy infrastructure-as-code at the scale of an AWS"
            " organization."
        ),
    )
    parser.add_argument(
        "-p",
        "--package-file",
        default=config.DEFAULT_PACKAGE_FILE,
        metavar="FILENAME",
        help=(
            "Location of the package definition YAML file."
            f" Default is {config.DEFAULT_PACKAGE_FILE}"
        ),
    )
    parser.add_argument(
        "-o",
        "--output-file",
        default=config.DEFAULT_OUTFILE_FILE,
        metavar="FILENAME",
        help=(
            "Location of the JSON file to which the command output details are written."
            f" Default is {config.DEFAULT_OUTFILE_FILE}"
        ),
    )
    parser.add_argument(
        "--temp-dir",
        default=config.DEFAULT_TEMP_DIR,
        metavar="DIRNAME",
        help=(
            "Location of the folder that stores cache and detailed log files. Default"
            f" is {config.DEFAULT_TEMP_DIR}"
        ),
    )
    parser.add_argument(
        "--force-orga-refresh",
        action="store_true",
        help=(
            "Ignore the cache in S3 and force the tool to query AWS Organizations for"
            " information on accounts and organizational unit"
        ),
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Increase log verbosity for debugging",
    )
    subparsers = parser.add_subparsers(title="command", dest="command")
    subparsers.required = True

    # Orga command
    subparsers.add_parser(
        "orga", help="Export AWS account list and organization structure"
    )

    # Detailed exit code
    parent_detailed_exitcode = argparse.ArgumentParser(add_help=False)
    parent_detailed_exitcode.add_argument(
        "--detailed-exitcode",
        action="store_true",
        help=(
            "Exit code is 0 if succeeded with no changes to be made, 1 if error, 2 if"
            " succeeded with changes present"
        ),
    )

    # Arguments that are common to the list, preview and apply commands
    parent_list_preview_apply = argparse.ArgumentParser(add_help=False)
    parent_list_preview_apply.add_argument(
        "-f",
        "--force-update",
        help=(
            "By default, modules are redeployed to a given AWS account and region only"
            " when the module hash or the variables changes. Add this argument to force"
            " module redeployment. Warning: this may result in a large number of"
            " deployments"
        ),
        action="store_true",
    )
    scope_group = parent_list_preview_apply.add_argument_group(
        title="filters",
        description=(
            "Restricts the deployment scope defined in the package configuration file,"
            " by excluding or including only certain modules, AWS accounts or regions."
        ),
    )
    scope_group.add_argument(
        "--include-modules",
        nargs="+",
        help="Include only certain modules",
        metavar="MODULE",
    )
    scope_group.add_argument(
        "--include-regions",
        nargs="+",
        help="Include only certain AWS regions",
        metavar="REGION",
    )
    scope_group.add_argument(
        "--include-account-ids",
        nargs="+",
        help="Include only certain AWS account IDs",
        metavar="ACCOUNT_ID",
        type=_account_id_type,
    )
    scope_group.add_argument(
        "--include-account-tags",
        nargs="+",
        help=(
            "Include only the AWS accounts with certain tags. Tags are cumulative, i.e."
            " KEY1=VALUE1 and KEY2=VALUE2 includes accounts with a tag KEY1 = VALUE1"
            " and with a tag TAG2 = VALUE2"
        ),
        metavar="KEY=VALUE",
        type=_tag_type,
    )
    scope_group.add_argument(
        "--include-account-names",
        nargs="+",
        help=(
            "Include only certain AWS account names. You can include wildcards (*) like"
            ' "*-prod"'
        ),
        metavar="ACCOUNT_NAME",
    )
    scope_group.add_argument(
        "--include-ou-ids",
        nargs="+",
        help="Include only certain organization unit IDs",
        metavar="OU_ID",
    )
    scope_group.add_argument(
        "--include-ou-tags",
        nargs="+",
        help=(
            "Include only the organizational units with certain tags. Tags are"
            " cumulative, i.e. KEY1=VALUE1 and KEY2=VALUE2 includes OUs with a tag KEY1"
            " = VALUE1 and with a tag TAG2 = VALUE2"
        ),
        metavar="KEY=VALUE",
        type=_tag_type,
    )
    scope_group.add_argument(
        "--exclude-modules", nargs="+", help="Exclude certain modules", metavar="MODULE"
    )
    scope_group.add_argument(
        "--exclude-regions",
        nargs="+",
        help="Exclude certain AWS regions",
        metavar="REGION",
    )
    scope_group.add_argument(
        "--exclude-account-ids",
        nargs="+",
        help="Exclude certain AWS account IDs",
        metavar="ACCOUNT_ID",
        type=_account_id_type,
    )
    scope_group.add_argument(
        "--exclude-account-tags",
        nargs="+",
        help=(
            "Exclude the AWS accounts with certain tags. Tags are cumulative, i.e."
            " KEY1=VALUE1 and KEY2=VALUE2 excludes accounts with a tag KEY1 = VALUE1"
            " and with a tag TAG2 = VALUE2"
        ),
        metavar="KEY=VALUE",
        type=_tag_type,
    )
    scope_group.add_argument(
        "--exclude-account-names",
        nargs="+",
        help=(
            "Exclude certain AWS account names. You can include wildcards (*) like"
            ' "*-prod"'
        ),
        metavar="ACCOUNT_NAME",
    )
    scope_group.add_argument(
        "--exclude-ou-ids",
        nargs="+",
        help="Exclude certain organization unit IDs",
        metavar="OU_ID",
    )
    scope_group.add_argument(
        "--exclude-ou-tags",
        nargs="+",
        help=(
            "Exclude the organizational units with certain tags. Tags are cumulative,"
            " i.e. KEY1=VALUE1 and KEY2=VALUE2 excludes OUs with a tag KEY1 = VALUE1"
            " and with a tag TAG2 = VALUE2"
        ),
        metavar="KEY=VALUE",
        type=_tag_type,
    )

    # List command
    subparsers.add_parser(
        "list",
        help="List deployed modules and deployments to create, update or destroy",
        parents=[parent_detailed_exitcode, parent_list_preview_apply],
    )

    # Arguments that are common to preview and apply commands
    parent_preview_apply = argparse.ArgumentParser(add_help=False)
    parent_preview_apply.add_argument(
        "--non-interactive",
        action="store_true",
        help="Do not ask to review and confirm the deployment scope",
    )
    parent_preview_apply.add_argument(
        "--keep-deployment-cache",
        action="store_true",
        help=(
            "Keep temporary files created during module deployment to enable"
            " troubleshooting"
        ),
    )

    # Arguments that are common to apply and update-hash
    parent_apply_update_hash = argparse.ArgumentParser(add_help=False)
    parent_apply_update_hash.add_argument(
        "--save-state-every-seconds",
        type=_check_positive_int,
        metavar="SECONDS",
        default=0,  # 0 means that package state is not saved periodically
        help=(
            "Save the package state periodically to S3 during execution to recover from"
            " an abrupt interruption. Specify a value in seconds larger than zero."
        ),
    )

    # Preview command
    subparsers.add_parser(
        "preview",
        help=(
            "Preview resources to add, update or delete when pending deployments are"
            " applied"
        ),
        parents=[
            parent_detailed_exitcode,
            parent_list_preview_apply,
            parent_preview_apply,
        ],
    )

    # Apply command
    subparsers.add_parser(
        "apply",
        help="Apply pending deployments",
        parents=[
            parent_detailed_exitcode,
            parent_list_preview_apply,
            parent_preview_apply,
            parent_apply_update_hash,
        ],
    )

    # Update-hash command
    subparsers.add_parser(
        "update-hash",
        help=(
            "Update the value of the module hash. This is useful to edit the"
            " module source code without needing to update deployments"
        ),
        parents=[
            parent_detailed_exitcode,
            parent_list_preview_apply,
            parent_preview_apply,
            parent_apply_update_hash,
        ],
    )

    # Arguments that are specific to the command "remove-orphans"
    remove_orphans = subparsers.add_parser(
        "remove-orphans",
        help=(
            "Remove orphaned module deployments from the package state corresponding to"
            " accounts that no longer exist in the AWS organization or regions that are"
            " no longer enabled in an account"
        ),
        parents=[parent_detailed_exitcode],
    )
    remove_orphans.add_argument(
        "--dry-run",
        action="store_true",
        help=(
            "Return the list of orphaned module deployments to remove without making"
            " any changes"
        ),
    )

    # Parse the arguments and store them as a dict for use by other modules
    config.CLI = vars(parser.parse_args())
    # Use the absolute path of the temporary folder
    config.CLI["temp_dir"] = path.abspath(config.CLI["temp_dir"])


def _create_cli_logger() -> logging.Logger:
    """Create and return a logger for the CLI that writes to stdout.

    Returns:
        CLI logger
    """
    logger = logging.getLogger("aws_orga_deployer")
    logger.setLevel(logging.DEBUG if config.CLI["debug"] is True else logging.INFO)
    stream_handler = logging.StreamHandler(sys.stdout)
    if config.CLI["debug"] is True:
        log_format = "%(levelname)s %(threadName)s:%(name)s:%(funcName)s %(message)s"
    else:
        log_format = "%(levelname)s %(message)s"
    formatter = logging.Formatter(log_format)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger


def init_cli() -> logging.Logger:
    """Main function to parse CLI arguments and initialize a logger.

    Returns:
        CLI logger
    """
    _parse_cli_args()
    return _create_cli_logger()



"""Constants or variables shared across multiple modules."""

# COMPLETED
from typing import Any, Dict

# CLI arguments populated by the module `cli`
CLI: Dict[str, Any] = {}

# Package configuration settings populated by the sub-package `package`
PACKAGE: Dict[str, Any] = {}

# Engine class instances for each module, populated by the module `Modules`
MODULES: Dict = {}

# Default location of the package definition YAML file
DEFAULT_PACKAGE_FILE: str = "package.yaml"

# Default location of the output JSON file
DEFAULT_OUTFILE_FILE: str = "output.json"

# Default location of the temporary directory that contains cache and logs
DEFAULT_TEMP_DIR: str = ".aws-orga-deployer"

# Number of concurrent threads used to query AWS Organizations
CONCURRENT_ORGA_THREADS: int = 10

# Name of the S3 file that stores the organization cache
ORGA_CACHE_FILENAME: str = "orga.json"

# Default period in seconds during which the organization description cached in
# S3 is reused instead of querying AWS Organizations (5 minutes)
DEFAULT_ORGA_CACHE_EXPIRATION: int = 5 * 60

# Name of the package state file in S3
STATE_FILENAME: str = "state.json"

# Name of the optional file in the module folder that contains patterns of
# filename to include or exclude to calculate module hash
HASH_CONFIG_FILENAME: str = "hash-config.json"

# Default number of concurrent workers
DEFAULT_CONCURRENT_WORKERS: int = 10

# Number of seconds during which the same AWS temporary credentials are provided
# to workers that must assume the same IAM role
CACHE_AWS_TEMP_CREDS: int = 5 * 60

# Duration of the AWS temporary credentials in seconds
AWS_TEMP_CREDS_DURATION: int = 60 * 60



"""Test the module `aws_orga_deployer.orga`."""

# COMPLETED
import unittest

from moto import mock_organizations, mock_s3, mock_sts

from aws_orga_deployer.orga import OrgaParser
from tests import mock
from tests.mock import mock_others


class TestOrgaParser(unittest.TestCase):
    """Test the class OrgaParser."""

    @mock_sts
    @mock_others
    def setUp(self):
        """Create mock resources and an OrgaParser object that must exist for
        all tests.
        """
        # Mock CLI arguments and package configuration settings
        mock.mock_cli_arguments(package_filename="package1.yaml")
        mock.mock_package_config()
        # Enable mocking for S3 and Organizations in all tests without having
        # to add declarators to each test function
        self.mock_s3 = mock_s3()
        self.mock_s3.start()
        self.mock_organizations = mock_organizations()
        self.mock_organizations.start()
        # Create a fake S3 bucket to store the organisation cache
        mock.create_fake_bucket()
        # Create a fake AWS organization
        self.fake_orga = mock.FakeOrgaParser()
        # Create an OrgaParser object that queries the fake organization
        self.orga = OrgaParser()

    def tearDown(self):
        """Disable mocking for S3 and Organizations."""
        self.mock_s3.stop()
        self.mock_organizations.stop()

    @mock_sts
    @mock_others
    def test_reload(self):
        """Check that creating an OrgaParser works with existing cache."""
        orga = OrgaParser()
        self.assertTrue(orga.loaded_from_cache)

    def test_fetch_accounts(self):
        """Check that the value of the internal attribute `accounts` which
        stores the list of accounts is correct.
        """
        self.assertDictEqual(
            self.orga.accounts,
            {
                "123456789012": {  # Default account ID in moto
                    "Name": "master",  # Default account ID in moto
                    "ParentOUs": [self.fake_orga.root_id],
                    "Tags": {},
                    "EnabledRegions": mock.ENABLED_REGIONS,
                },
                self.fake_orga.account_a_id: {
                    "Name": "account-a-override",
                    "ParentOUs": [self.fake_orga.root_id],
                    "Tags": {"Name": "account-a-override", "Environment": "prod"},
                    "EnabledRegions": mock.ENABLED_REGIONS,
                },
                self.fake_orga.account_b_id: {
                    "Name": "account-b",
                    "ParentOUs": [self.fake_orga.ou_prod_id, self.fake_orga.root_id],
                    "Tags": {"Environment": "prod"},
                    "EnabledRegions": mock.ENABLED_REGIONS,
                },
                self.fake_orga.account_c_id: {
                    "Name": "account-c",
                    "ParentOUs": [self.fake_orga.ou_test_id, self.fake_orga.root_id],
                    "Tags": {"Environment": "test"},
                    "EnabledRegions": mock.ENABLED_REGIONS,
                },
            },
        )

    def test_fetch_ous(self):
        """Check that the value of the internal attribute `_ous` is correct."""
        self.assertDictEqual(
            self.orga.ous,
            {
                self.fake_orga.root_id: {"Tags": {}, "Name": "root"},
                self.fake_orga.ou_prod_id: {
                    "Tags": {"Environment": "prod"},
                    "Name": "prod",
                },
                self.fake_orga.ou_test_id: {
                    "Tags": {"Environment": "test"},
                    "Name": "test",
                },
            },
        )

    def test_get_all_accounts(self):
        """Check that `test_get_all_accounts` returns all accounts."""
        self.assertCountEqual(
            self.orga.get_all_accounts(),
            [
                "123456789012",
                self.fake_orga.account_a_id,
                self.fake_orga.account_b_id,
                self.fake_orga.account_c_id,
            ],
        )

    def test_get_accounts_by_id(self):
        """Check that `get_accounts_by_id` returns only account IDs that exist."""
        self.assertCountEqual(
            self.orga.get_accounts_by_id(
                [
                    self.fake_orga.account_a_id,
                    self.fake_orga.account_b_id,
                    "098765432109",
                ],  # Fake account ID
            ),
            [
                self.fake_orga.account_a_id,
                self.fake_orga.account_b_id,
            ],
        )

    def test_get_accounts_by_name(self):
        """Check that `get_accounts_by_name(["*-override"])` only returns the
        ID of account A.
        """
        self.assertCountEqual(
            self.orga.get_accounts_by_name(["*-override"]),
            [self.fake_orga.account_a_id],
        )

    def test_get_accounts_by_ou(self):
        """Check that `get_accounts_by_ou([root_id])` returns the ID of all
        accounts in the organization.
        """
        self.assertCountEqual(
            self.orga.get_accounts_by_ou([self.fake_orga.root_id]), self.orga.accounts
        )

    def test_get_accounts_by_tag(self):
        """Check that `get_accounts_by_tag(["Environment=prod"])` returns both
        accounts that have this tag.
        """
        self.assertCountEqual(
            self.orga.get_accounts_by_tag(["Environment=prod"]),
            [
                self.fake_orga.account_a_id,
                self.fake_orga.account_b_id,
            ],
        )

    def test_get_accounts_by_ou_tag(self):
        """Check that `get_accounts_by_ou(["Environment=test"])` returns only
        the account C.
        """
        self.assertCountEqual(
            self.orga.get_accounts_by_ou_tag(["Environment=test"]),
            [self.fake_orga.account_c_id],
        )

    def test_get_account_regions_1(self):
        """Check that `get_account_regions(account_id, [`ALL_ENABLED])` returns
        all enabled regions for an account.
        """
        self.assertCountEqual(
            self.orga.get_account_regions(self.fake_orga.account_a_id, ["ALL_ENABLED"]),
            mock.ENABLED_REGIONS,
        )

    def test_get_account_regions_2(self):
        """Check that `get_account_regions(account_id, REGIONS)` returns only
        the regions that intersect with the enabled regions.
        """
        self.assertCountEqual(
            self.orga.get_account_regions(
                self.fake_orga.account_a_id,
                ["us-east-1", "eu-west-1", "eu-west-3"],
            ),
            ["us-east-1", "eu-west-1"],
        )

    def test_get_all_enabled_regions(self):
        """Check that `get_all_enabled_regions()` returns all enabled regions."""
        self.assertCountEqual(
            self.orga.get_all_enabled_regions(),
            mock.ENABLED_REGIONS,
        )

    def test_get_account_name(self):
        """Check that get_account_name returns the expected account name."""
        self.assertEqual(
            self.orga.get_account_name(self.fake_orga.account_a_id),
            "account-a-override",
        )

    def test_get_account_name_invalid(self):
        """Check that get_account_name returns "undefined" if the account ID
        does not exist.
        """
        self.assertEqual(
            self.orga.get_account_name("098765432109"),
            "undefined",
        )

    def test_account_region_exists(self):
        """Check that account_region_exists returns True given that the account
        exists and the region is enabled in this account.
        """
        self.assertTrue(self.orga.account_region_exists("123456789012", "us-east-1"))

    def test_account_region_not_exists(self):
        """Check that account_region_exists returns False given that the account
        exists and the region is not enabled in this account.
        """
        self.assertFalse(self.orga.account_region_exists("123456789012", "eu-west-3"))
